%%%----------------------------------------------------------------------
%%% File    : hipe_assign.erl
%%% Author  : Ingemar berg <d95ina@zeppo.it.uu.se>
%%% Purpose : Find out what color each live range shall have.
%%% Created : 18 Mar 2000 by Ingemar berg <d95ina@zeppo.it.uu.se>
%%%----------------------------------------------------------------------

-module(hipe_assign).
-export([assignColors/6, defaultColoring/4]).

%%%----------------------------------------------------------------------
% Function:    assignColors
%
% Description: Tries to assign colors to nodes in a stack.
% Parameters:
%   Stack          --  The SelectStack built by the Select function, 
%                      this stack contains tuples in the form {Node,Edges} 
%                      where  Node is the Node number and Edges is an ordset 
%                      containing the numbers of all the adjacent nodes.
%   NodeSets       --  This is a record containing all the different node 
%                      sets that are used in the register allocator.
%   Alias          --  This is a mapping from nodes to nodes, if a node has 
%                      been coalesced this mapping shows the alias for that 
%                      node.
%   AllColors      --  This is an ordset containing all the available colors
%
%   Target         --  The module containing the target-specific functions.
%
% Returns:
%   Color          --   A mapping from nodes to their respective color.
%   NodeSets       --   The updated node sets
%                 
%%%----------------------------------------------------------------------

assignColors (Stack,NodeSets,Color,Alias,AllColors, Target) ->
    case Stack of
	[] ->
	    {Color,NodeSets};
	[{Node,Edges}|Stack1] ->
	    case hipe_node_sets:member(precolored,Node, NodeSets) of
		true ->                                 % Allready colored
		    hipe_message:debug(assign, "Node ~p allready colored~n",[Node]),
		    assignColors(Stack1,NodeSets,Color,Alias,AllColors,Target);
		false ->                                % Try to find color
		    OkColors = findOkColors(Edges,AllColors,Color,NodeSets,
					    Alias),
		    case OkColors of
			[] ->                           % Spill case
			    NodeSets1 = hipe_node_sets:add(spilled,Node,NodeSets),
			    assignColors(Stack1,NodeSets1,Color,
					 Alias,AllColors,Target);
			[Col|Cols] ->                   % Colorize case
			    NodeSets1 = hipe_node_sets:add(colored, Node,NodeSets),
			    Color1 = 
				hipe_color:setColor(Node,
						    Target:physical_name(Col),
					            Color),
			    assignColors(Stack1,NodeSets1,Color1,Alias,
					 AllColors,Target)
		    end
	    end
    end.

%%%---------------------------------------------------------------------
% Function:     defaultColoring
% 
% Description: Make the default coloring
% Parameters:
%   Regs           -- The list of registers to be default colored
%   Color          -- The color mapping that shall be changed
%   NodeSets       -- The node sets that shall be updated
%   Target         -- The module containing the target-specific functions.
%
% Returns:
%   NewColor       -- The updated color mapping
%   NewNodeSets    -- The updated node sets
%
%%%---------------------------------------------------------------------

defaultColoring([],Color,NodeSets, Target) ->
    {Color,NodeSets};
defaultColoring([Reg|Regs],Color,NodeSets, Target) ->
    Color1 = hipe_color:setColor(Reg,Target:physical_name(Reg),Color),
    NodeSets1 = hipe_node_sets:add(colored,Reg,NodeSets),
    defaultColoring(Regs,Color1,NodeSets1, Target).
  
%%% Map a function to a list (sort of like in for example SML).

map (F,[]) ->
    [];
map (F,[X|Xs]) -> 
    [apply(F,[X])|map(F,Xs)].

%%% Find the colors that are OK for a node with certain edges.

findOkColors(Edges,AllColors,Color,NodeSets,Alias) ->
    F = fun (Node) ->
		hipe_alias:getAlias(Node,NodeSets,Alias) 
	end,
    Edges2 = map(F,Edges),
    find(Edges2,AllColors,Color,NodeSets,Alias).

%%% Find all the colors of the nodes in the list [Node|Nodes] and remove them 
%%% from the set OkColors, when the list is empty, return OkColors.

find([],OkColors,Color,NodeSets,Alias) ->
    ordsets:to_list(OkColors); 
find([Node|Nodes],OkColors,Color,NodeSets,Alias) ->
    case (hipe_node_sets:member(colored,Node,NodeSets) or 
	  hipe_node_sets:member(precolored,Node,NodeSets)) of
	true ->
	    Col = hipe_color:getColor(Node,Color),
	    OkColors1 = ordsets:del_element(Col,OkColors),
	    find(Nodes,OkColors1,Color,NodeSets,Alias);
	false ->
	    find(Nodes,OkColors,Color,NodeSets,Alias)
    end.
