%%%----------------------------------------------------------------------
%%% File    : hipe_reg_sel.erl
%%% Author  : Ingemar berg <d95ina@zeppo.it.uu.se>
%%% Purpose : Spill nodes that can't be dealt with in any other way
%%% Created : 20 Mar 2000 by Ingemar berg <d95ina@zeppo.it.uu.se>
%%%----------------------------------------------------------------------

-module(hipe_reg_sel).
-author('d95ina@zeppo.it.uu.se').
-export([selectSpill/7]).

%%%---------------------------------------------------------------------
% Function: selectSpill
% 
% Description:     Select the node to spill and spill it
% Parameters:
%   WorkLists      -- A datatype containing the different worklists
%   Moves          -- A datatype containing the move sets
%   IG             -- The interference graph
%   K              -- The number of available registers
%   NodeSets       -- A datatype containing the node sets
%   Alias          -- The alias mapping
%   SpillLimit     -- Try not to spill any nodes above the spill limit
%
% Returns:
%   WorkLists      -- The updated worklists
%   Moves          -- The updated moves
%
%%%---------------------------------------------------------------------

selectSpill(WorkLists, Moves, IG, K, NodeSets, Alias, SpillLimit) ->
    [CAR|CDR] = hipe_reg_worklists:spill(WorkLists),
    SpillCost = getCost(CAR, IG,SpillLimit),
    M = findCheapest(CDR,IG,SpillCost,CAR, SpillLimit),
    WorkLists1 = hipe_reg_worklists:remove(spill,M,WorkLists),
    WorkLists2 = hipe_reg_worklists:add(simplify,M,WorkLists1),
    hipe_freeze:freezeMoves(M, K, WorkLists2, Moves,
			    hipe_ig:degree(IG), NodeSets, Alias).

%%% Find the node that is cheapest to spill

findCheapest([], IG, Cost, Cheapest, SpillLimit) ->
    Cheapest;
findCheapest([Node|Nodes], IG, Cost, Cheapest, SpillLimit) ->
    ThisCost = getCost(Node, IG, SpillLimit),
    case ThisCost <  Cost of
	true ->
	    findCheapest(Nodes, IG, ThisCost, Node, SpillLimit);
	false ->
	    findCheapest(Nodes, IG, Cost, Cheapest, SpillLimit)
    end.

%%% Get the cost for spilling a certain node, node numbers above the spill 
%%% limit are extremely expensive.

getCost(Node, IG, SpillLimit) ->
  case Node > SpillLimit of
    true ->  inf;
    false -> hipe_spillcost:spill_cost(Node, IG)
  end.
