%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%		  GRAPH COLORING REGISTER ALLOCATOR
%
% A simple graph coloring register allocator. Now with 
% iterative register coalescing:
%
% - build interference graph + estimate spill costs
% - simplify graph, coalesce move related nodes, repeat, freeze, repeat
% - select colors
%
% Emits a coloring: a list of {TempName,Location}
%  where Location is {reg,N} or {spill,M}.
%   and {reg,N} denotes some register N
%   and {spill,M} denotes the Mth spilled node
%
% This version uses vectors rather than hash tables, and uses
% faster algorithms since all vars are known at the start.
%   The result should be considerably quicker than earlier versions.
%
% be considerably faster as a result.
% Deficiencies:
% - no iterated graph coloring    (means more registers must be reserved)
%   * requires code rewriting to work
% - no rewriting of code          (spills, etc.)
% - no renaming                   (reduce unnecessary register pressure)
% - spill costs are naive         (should use better exec.estimates)
% - no live range splitting       (possibly not critical)
%


-module(hipe_irc_regalloc).
-export([regalloc/3]).

-define(debug(X,Y), io:format(X, Y)).
%-define(debug(X,Y), ok).

-record(node, {adj=[], degree=0, moves=[]}).

%
% Given CFG produce a coloring list of items {reg,N} (1 =< N =< K) 
% and {spill,M}, where M is an index denoting 'a location'. (might 
% be uses as a stack index, perhaps.)
%

regalloc(CFG, SpillIx, SpillLimit) ->
   PhysRegs = hipe_sparc_registers:allocatable(),
   ?debug("building IG~n", []),
   IG = build(CFG),
   {Cols, SpillIx0} = color(IG, ordsets:from_list(PhysRegs), SpillIx),
   %% check_coloring(Cols, IG),
   {Cols, SpillIx0}.



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% *** BUILDING THE INTERFERENCE GRAPH ***
%
% Returns {Interference_graph, Spill_cost_dictionary}
%

build(CFG) ->
   Live = hipe_sparc_liveness:analyze(CFG),
   {_, NumN} = hipe_sparc_cfg:var_range(CFG),
   Labels = hipe_sparc_cfg:labels(CFG),
   IG = empty_ig(NumN),
   IG0 = build_bbs(Labels, CFG, Live, IG),
   normalize_ig(IG0).

build_bbs([], CFG, Live, IG) ->
   IG;
build_bbs([L|Ls], CFG, Live, IG) ->
   Code = hipe_bb:code(hipe_sparc_cfg:bb(CFG, L)),
   LiveOut = reg_names(ordsets:to_list(hipe_sparc_liveness:liveout(Live, L))),
   {_, NewIG} = build_bb(Code, LiveOut, IG),
   build_bbs(Ls, CFG, Live, NewIG).

build_bb([], LiveOut, IG) ->
   {LiveOut, IG};
build_bb([X|Xs], LiveOut, IG) ->
   {Live, NewIG} = build_bb(Xs, LiveOut, IG),
   build_instr(X, Live, NewIG).

% Note: We could add move-related arcs here as well.
%
% Note: Ideally, we would like to add all registers to the IG
%  at once rather than doing 'add_nodes' for each instruction.
% (This is costly, since nodes that already are present are checked!)

build_instr(I, Live, IG) ->
   {Def0, Use0} = hipe_sparc:def_use(I),
   Def = ordsets:from_list(reg_names(Def0)),
   Use = ordsets:from_list(reg_names(Use0)),
   case {hipe_sparc:is_move(I), Use, Def} of
      {true, [X], [X]} ->
	 Live0 = Live,
	 IG0 = IG;
      {true, [U], [D]} ->
	 Live0 = ordsets:del_element(U, Live),
	 IG0 = add_ig_moves(IG, U, D);
      _ ->
	 Live0 = Live,
	 IG0 = IG
   end,
   IG1 = interference_arcs(Def, Live0, IG0),
   Live1 = ordsets:union(Use, ordsets:subtract(Live0, Def)),
   {Live1, IG1}.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

interference_arcs([], Live, IG) ->
   IG;
interference_arcs([X|Xs],Live,IG) ->
   interference_arcs(Xs, Live, i_arcs(X, Live, IG)).

i_arcs(X, [], IG) ->
   IG;
i_arcs(X, [Y|Ys], IG) ->
   i_arcs(X, Ys, add_edge(X, Y, IG)).



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% *** COLORING ***
%
% Coloring is done straightforwardly:
% - find the low-degree nodes, put them in low
% - while low non-empty:
%   * remove x from low
%   * push x on stack
%   * decrement degree of adj of x
%   * for each neighbor y of low degree, put y on low
% - when low empty:
%   - if graph empty, return stack
%   - otherwise
%     * select a node z to spill
%     * push z on stack
%     * decrement degree of adj of z
%     * add low-degree adj of z to low
%     * restart the while-loop above

color(IG, PhysRegs, SpillIx) ->
   K = ordsets:size(PhysRegs),
   Stack = [],
   Coalesced = [],
   Alias = empty_alias(),
   {Lows, Moves, Left, PreC} = mk_worklists(IG, K),
   {Stack0, IG0, Coalesced0} =
      simplify_coalesce_freeze(Stack, IG, Lows, Moves, Coalesced, Alias, 
			       K, Left),
   select(Stack0, Coalesced0, IG0, K, PhysRegs, Left+PreC, SpillIx).


mk_worklists(IG, K) ->
   Nodes = list_ig(IG),
   mk_worklists0(Nodes, K, [], ordsets:new(), 0, 0).

mk_worklists0([], K, Lows, Moves, Left, PreC) ->
   {Lows, Moves, Left, PreC};
mk_worklists0([{N,Info}|Nodes], K, Lows, Moves, Left, PreC) ->
   case hipe_sparc_registers:is_precolored(N) of
      true ->
	 PreC0 = PreC+1,
	 Left0 = Left;
      false ->
	 PreC0 =  PreC,
	 Left0 = Left+1
   end,
   AdjMoves = Info#node.moves,
   if AdjMoves =:= [] ->
	 if Info#node.degree < K ->
	       mk_worklists0(Nodes, K, [N|Lows], Moves, Left0, PreC0);
	    true ->
	       mk_worklists0(Nodes, K, Lows, Moves, Left0, PreC0)
	 end;
      true ->
	 MoreMoves = mk_moves(N, AdjMoves, ordsets:new()),
	 Moves0 = ordsets:union(Moves, MoreMoves),
	 mk_worklists0(Nodes, K, Lows, Moves0, Left0, PreC0)
   end.

mk_moves(N, [], Moves) ->
   Moves;
mk_moves(N, [To|Tos], Moves) ->
   if N < To ->
	 mk_moves(N, Tos, ordsets:add_element({N, To}, Moves));
      true ->
	 mk_moves(N, Tos, ordsets:add_element({To, N}, Moves))
   end.



simplify_coalesce_freeze(Stack, IG, Lows, Moves, Coalesced, Alias, K, Left) ->
   ?debug("Left: ~w~n", [Left]),
   {Stack0, IG0, Left0} = simplify(Lows, Stack, IG, K, Left),
   ?debug("Left0: ~w~n", [Left0]),
   {IG1, Coalesced0, Lows0, Moves0, Alias0} = 
      coalesce(IG0, ordsets:to_list(Moves), K, Coalesced, Alias),
   Left1 = Left0 - length(Coalesced0)+length(Coalesced),
   ?debug("Left1: ~w~n", [Left1]),
   if Left1 =:= 0 ->
	 {Stack0, IG1, aliasify(Alias0, Coalesced0)};
      Lows0 =:= [] ->
	 {IG2, Lows1, Moves1} = freeze(IG1, Moves0, K),
	 ?debug("Ready: ~w~n", [Lows1]),
	 if Lows1 =:= [] ->
	       IG3 = freeze_all(IG2, Moves1),
	       {Spilled, Lows2, IG4} = spill(IG3, K, Coalesced0),
	       ?debug("Spilled: ~w~n", [Spilled]),
	       Stack1 = push_spilled(Stack0, Spilled),
	       ?debug("Pushed: ~w~n", [length(Stack1)]),
	       Left2 = Left1 - length(Spilled),
	       simplify_coalesce_freeze(Stack1, IG4, Lows2, [], Coalesced0,
					Alias0, K, Left2);
	    true ->
	       simplify_coalesce_freeze(Stack0, IG2, Lows1, Moves1, Coalesced0,
					Alias0, K, Left1)
	 end;
      true ->
	 simplify_coalesce_freeze(Stack0, IG1, Lows0, Moves0, Coalesced0,
				  Alias0, K, Left1)
   end.



aliasify(Alias, []) ->
   [];
aliasify(Alias, [{To, From} | Coalesced]) ->
   [{get_alias(Alias, To), From} | aliasify(Alias, Coalesced)].


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Simplification: push all easily colored nodes on a stack;
%  when the list of easy nodes becomes empty, see if graph is
%  empty as well. If it is not, spill a node and continue.
%  If it is empty, return the stack.
%
% Notes:
% - We keep the set of visited nodes around for spill purposes
%   (visited nodes are not considered for spilling)
%
% - Arguments:
%   Low: low-degree nodes (ready to color)
%   NumNodes: number of remaining nodes in graph
%   IG: interference graph
%   Spill: spill costs of nodes
%   K: number of colors
%   Ix: next spill index
%   Stk: stack of already simplified nodes
%

simplify([], Stack, IG, K, Left) ->
   {Stack, IG, Left};
simplify([Low|Lows], Stack, IG, K, Left) ->
   {Lows0, IG0} = decrement_adj(Low, Lows, IG, K),
   Stack0 = push_colored(Stack, Low),
   simplify(Lows0, Stack0, IG0, K, Left-1).


%
% Returns { NowLowDegreeAdj, NewIG }
%

decrement_adj(Node, Lows, IG, K) ->
   Adj = adj(Node, IG),
   decrement_each(Adj, Lows, IG, K).


% For each node, decrement its degree and check if it is now
% a low-degree node. In that case, add it to the 'low list'.

decrement_each([], Low, IG, K) ->
   {Low, IG};
decrement_each([A|Adj], Low, IG, K) ->
   {Low0, IG0} = decrement_each(Adj, Low, IG, K),
   Info = hipe_vectors:get(IG0, A),
   Degree = Info#node.degree,
   Moves = Info#node.moves,
   IG1 = hipe_vectors:set(IG0, A, Info#node{degree=Degree-1}),
   if
      Degree =:= K, Moves =:= [] ->
	 {[A|Low0], IG1};
      true ->
	 {Low0, IG1}
   end.



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Coalesce nodes
%
%   The size of the union of the two nodes adj must be < K
%

coalesce(IG, Moves, K, Coalesced, Alias) ->
   ?debug("Moves: ~w~n", [Moves]),
   coalesce(Moves, IG, K, Coalesced, [], [], Alias).

coalesce([], IG, K, Coalesced, Lows, MovesLeft, Alias) ->
    ?debug("Coalesced: ~w~n", [Coalesced]),
    ?debug("MovesLeft: ~w~n", [MovesLeft]),
    Lows0 = ordsets:from_list(rename_lows(ordsets:from_list(Lows),
					  Alias)),
    Lows1 = add_lows(Lows0, [], IG, K),
    ?debug("Ready: ~w~n", [Lows1]),
    {IG, Coalesced, Lows1, ordsets:from_list(MovesLeft), Alias};
coalesce([{X, Y}|Ms], IG, K, Coalesced, Lows, MovesLeft, Alias) ->
   Xp = get_alias(Alias, X),
   Yp = get_alias(Alias, Y),
   case hipe_sparc_registers:is_precolored(Yp) of
      true ->
	 U = Yp,
	 V = Xp;
      false ->
	 U = Xp,
	 V = Yp
   end,
   case hipe_sparc_registers:is_precolored(V) or ordsets:is_element(U, adj(V, IG)) of
      true ->
	 ?debug("Constrained: ~w, (~w)~n", [{U, V}, {X, Y}]),
	 IG1 = unrelate_move(IG, U, V),
	 MovesLeft0 = lists:delete(list_to_tuple(lists:sort([U, V])),
				   MovesLeft),
	 coalesce(Ms, IG1, K, Coalesced, [U, V | Lows], MovesLeft0, Alias);
      false ->
	 AdjU = adj(U, IG),
	 AdjV = adj(V, IG),
	 Cond = case hipe_sparc_registers:is_precolored(U) of
		   true ->
		      ok(U, AdjV, IG, K);
		   false ->
		      conservative(ordsets:union(AdjU, AdjV), IG, K)
		end,
	 if Cond =:= true ->
	       ?debug("Combining: ~w~n", [{U, V}]),
	       IG0 = combine(IG, U, V, AdjU, AdjV, K),
	       Lows0 = [U | Lows],
	       Alias0 = add_alias(Alias, V, U),
	       Coalesced0 = [{U, V} | Coalesced],
	       coalesce(Ms, IG0, K, Coalesced0, Lows0, MovesLeft, Alias0);
	    true ->
	       ?debug("Not ready: ~w~n", [{X, Y}]),
	       coalesce(Ms, IG, K, Coalesced, Lows, [{X, Y}|MovesLeft], Alias)
	 end
   end.



add_lows([], Lows, IG, K) ->
   Lows;
add_lows([X|Xs], Lows, IG, K) ->
   Deg = degree(X, IG),
   Moves = moves(X, IG),
   if Deg < K, Moves =:= [] ->
	 add_lows(Xs, [X|Lows], IG, K);
      true ->
	 add_lows(Xs, Lows, IG, K)
   end.



ok(R, [], IG, K) ->
   true;
ok(R, [T|Ts], IG, K) ->
   ok0(R, T, IG, K) and ok(R, Ts, IG, K).

ok0(R, T, IG, K) ->
   case hipe_sparc_registers:is_precolored(T) of
      true ->
	 true;
      false ->
	 Deg = degree(T, IG),
	 if Deg < K ->
	       true;
	    true ->
	       AdjT = adj(T, IG),
	       ordsets:is_element(R, AdjT)
	 end
   end.



conservative(Nodes, IG, K) ->
   conservative(Nodes, IG, K, 0).

conservative([], IG, K, S) ->
   S < K;
conservative([N|Ns], IG, K, S) ->
   Deg = degree(N, IG),
   if Deg < K ->
	 conservative(Ns, IG, K, S);
      true ->
	 conservative(Ns, IG, K, S+1)
   end.



rename_lows([], Alias) ->
   [];
rename_lows([L|Ls], Alias) ->
   [get_alias(Alias, L) | rename_lows(Ls, Alias)].


get_alias(Alias, X) ->
    case gb_trees:lookup(X, Alias) of
	{value, Y} ->
	    get_alias(Alias, Y);
	none ->
	    X
    end.


add_alias(Alias, From, To) ->
   gb_trees:insert(From, To, Alias).


empty_alias() ->
   gb_trees:empty().




%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Freeze. When no simplification or coalescing can be done 'freeze' a
% move related node (of low degree) making it (and possibly other
% move related nodes) ready for simplification.
%

freeze(IG, Moves, K) ->
   Nodes = list_ig(IG),
   case low_degree_move_node(IG, K, Moves) of
      {none, _} ->
	 {IG, [], Moves};
      {Node, Moves0} ->
	 ?debug("Freezing: ~w~n", [Node]),
	 Related = moves(Node, IG),
	 IG1 = unrelate_moves(IG, Node, Related),
	 {IG1, add_lows([Node | Related], [], IG1, K), Moves0}
   end.


freeze_all(IG, []) ->
   IG;
freeze_all(IG, [{X, Y} | Moves]) ->
   InfoX = hipe_vectors:get(IG, X),
   IG1 = hipe_vectors:set(IG, X, InfoX#node{moves=[]}),
   InfoY = hipe_vectors:get(IG, Y),
   IG2 = hipe_vectors:set(IG1, Y, InfoY#node{moves=[]}),
   freeze_all(IG2, Moves).


low_degree_move_node(IG, K, []) ->
   {none, []};
low_degree_move_node(IG, K, [{X, Y}|Moves]) ->
   DegX = degree(X, IG),
   DegY = degree(Y, IG),
   if DegX < K ->
	 {X, Moves};
      DegY < K ->
	 {Y, Moves};
      true ->
	 {Node, Moves0} = low_degree_move_node(IG, K, Moves),
	 {Node, [{X, Y}|Moves0]}
   end.


unrelate_moves(IG, Node, []) ->
   Info = hipe_vectors:get(IG, Node),
   hipe_vectors:set(IG, Node, Info#node{moves=[]});
unrelate_moves(IG, Node, [X|Xs]) ->
   InfoX = hipe_vectors:get(IG, X),
   MovesX = ordsets:del_element(Node, InfoX#node.moves),
   IG0 = hipe_vectors:set(IG, X, InfoX#node{moves=MovesX}),
   unrelate_moves(IG0, Node, Xs).


% removes X from Y's movelist and vice versa

unrelate_move(IG, X, Y) ->
   InfoY = hipe_vectors:get(IG, Y),
   MovesY = ordsets:del_element(X, InfoY#node.moves),
   IG0 = hipe_vectors:set(IG, Y, InfoY#node{moves=MovesY}),
   InfoX = hipe_vectors:get(IG, X),
   MovesX = ordsets:del_element(Y, InfoX#node.moves),
   hipe_vectors:set(IG0, X, InfoX#node{moves=MovesX}).




%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% The spill cost of a node is:
%    est_spill_cost / current_degree
%
% For all unvisited nodes, compute spill cost and select the minimum.
% This node is chosen to be spilled. Then decrement the degree of its
% adj, and return those of low degree.
%
% Notes:
% - A better method for computing spill costs is to just keep the
%   minimum cost node. But for debugging purposes, we compute a list
%   of {node,spillcost} pairs and select the minimum.
%
% Returns:
%  {Spilled_node, Low_degree_adj, New_interference_graph}

spill(IG, K, Coalesced) ->
   Candidates = spill_candidates(IG, K, Coalesced),
   ?debug("Spill candidates: ~w~n", [Candidates]),
   Costs = spill_costs(IG, Candidates),
   ActualCosts = lists:reverse(lists:keysort(1, Costs)),
   spill_loop(IG, K, ActualCosts, []).

spill_loop(IG, K, [], Spilled) ->
   error(spill, "there is no node to spill");
spill_loop(IG, K, [{Cost,N}|ActualCosts], Spilled) ->
   {Low, IG0} = decrement_adj(N, [], IG, K),
   IG1 = set_degree(IG0, N, 0),
   ?debug("Spilled node ~w (~w now ready)~n", [N, Low]),
   case Low of
      [] when ActualCosts =/= [] ->
	 spill_loop(IG1, K, ActualCosts, [N|Spilled]);
      _ ->
	 {[N|Spilled], Low, IG1}
   end.


spill_costs(IG, []) ->
   [];
spill_costs(IG, [N|Ns]) ->
   case hipe_sparc_registers:is_precolored(N) of
      true ->
	 spill_costs(IG, Ns);
      false ->
	 [{degree(N, IG), N} | spill_costs(IG, Ns)]
   end.


spill_candidates(IG, K, Coalesced) ->
   Nodes = list_ig(IG),
   spill_candidates0(Nodes, K, Coalesced).

spill_candidates0([], K, Coalesced) ->
   [];
spill_candidates0([{N, Info}|Ns], K, Coalesced) ->
   case {Info#node.degree >= K,
	 hipe_sparc_registers:is_precolored(N),
	 is_coalesced(N, Coalesced)} of
      {true, false, false} ->
	 [N | spill_candidates0(Ns, K, Coalesced)];
      _ ->
	 spill_candidates0(Ns, K, Coalesced)
   end.


is_coalesced(N, []) ->
   false;
is_coalesced(N, [{_,N}|_]) ->
   true;
is_coalesced(N, [_|Coalesced]) ->
   is_coalesced(N, Coalesced).




%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Returns a list of {Name,Location}, where Location is
%   either {spill,M} or {reg,R}
%
% Note: we use pessimistic coloring here.
% - we could use optimistic coloring: for spilled node, check if there is
%   an unused color among the adj and choose that.

select(Stack, Coalesced, IG, K, PhysRegs, NumNodes, SpillIx) ->
   PreC = hipe_sparc_registers:all_precolored(),
   {PhysColors, Cols} = precolor(PreC, none_colored(NumNodes)),
   {Colors, SpillIx0} = select_colors(Stack, IG, Cols, PhysRegs, K, [], 
				      SpillIx),
   Colors0 = PhysColors ++ Colors,
   ?debug("Coalesced: ~w~n", [lists:keysort(1, Coalesced)]),
   ?debug("Colors: ~w~n", [lists:keysort(1, Colors0)]),
   Colors1 = add_coalesced_colors(lists:keysort(1, Colors0), 
				  lists:keysort(1, Coalesced)),
   {Colors1, SpillIx0}.


select_colors([], IG, Cols, PhysRegs, K, Colors, SpillIx) ->
   {Colors, SpillIx};
select_colors([{X,colorable}|Xs], IG, Cols, PhysRegs, K, Colors, SpillIx) ->
   {R,NewCols} = select_color(X, IG, Cols, PhysRegs),
   select_colors(Xs, IG, NewCols, PhysRegs, K, [{X,{reg,R}}|Colors], SpillIx);
select_colors([{X,spill}|Xs], IG, Cols, PhysRegs, K, Colors, SpillIx) ->
   select_colors(Xs, IG, Cols, PhysRegs, K, [{X,{spill,SpillIx}}|Colors], 
		 SpillIx+1).

select_color(X, IG, Cols, PhysRegs) ->
   K = ordsets:size(PhysRegs),
   Ns = adj(X, IG),
   UsedColors = get_colors(adj(X, IG), Cols),
   Reg = select_unused_color(UsedColors, PhysRegs),
   {Reg, set_color(X, Reg, Cols)}.



add_coalesced_colors([], []) ->
   [];
add_coalesced_colors([{Var, Color}|Colors], [{Var, From}|Coalesced]) ->
   [{From, Color} | add_coalesced_colors([{Var, Color}|Colors], Coalesced)];
add_coalesced_colors([Color|Colors], Coalesced) ->
   [Color|add_coalesced_colors(Colors, Coalesced)].


%%%%%%%%%%%%%%%%%%%%

get_colors([],Cols) -> [];
get_colors([X|Xs],Cols) ->
   case color_of(X,Cols) of
      uncolored ->
	 get_colors(Xs,Cols);
      {color,R} ->
	 [R|get_colors(Xs,Cols)]
   end.

select_unused_color(UsedColors, PhysRegs) ->
   Summary = ordsets:from_list(UsedColors),
   AvailRegs = ordsets:subtract(PhysRegs, Summary),
   %% hd(AvailRegs).
   select_avail_reg(AvailRegs).

% We choose the register to use randomly from the set of available
% registers. 
%
% Note: Another way of doing it is LRU-order:
% - Have an LRU-queue of register names; when coloring, try the colors in that
%   order (some may be occupied).
% - When a color has been selected, put it at the end of the LRU.

select_avail_reg(Regs) ->
   case get(seeded) of
      undefined ->
	 random:seed(),
	 put(seeded,true);
      true ->
	 ok
   end,
   NReg = length(Regs),
   RegNo = random:uniform(NReg),
   nth(RegNo, Regs).

nth(1,[R|_]) -> R;
nth(N,[_|Rs]) when N > 1 ->
   nth(N-1,Rs).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

push_spilled(Stack, []) ->
   Stack;
push_spilled(Stack, [X|Xs]) ->
   push_spilled([{X,spill}|Stack], Xs).

push_colored(Stack, X) ->
   ?debug("Pushed: ~w~n", [X]),
   [{X, colorable} | Stack].




%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% *** ABSTRACT DATATYPES ***


member(X,[]) -> false;
member(X,[X|_]) -> true;
member(X,[_|Ys]) -> member(X,Ys).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% The ig datatype:
%
% Note: if we know the number of temps used, we can use a VECTOR
% instead, which will speed up things.
%


empty_ig(NumNodes) ->
   hipe_vectors:init(NumNodes,#node{adj=[],degree=0, moves=[]}).

degree(Info) ->
   Info#node.degree.

moves(Info) ->
   Info#node.moves.

adj(Info) ->
   Info#node.adj.

add_ig_moves(IG, X, Y) ->
   InfoX = hipe_vectors:get(IG, X),
   OldXMoves = InfoX#node.moves,
   NewInfoX = InfoX#node{moves=[Y|OldXMoves]},
   IG0 = hipe_vectors:set(IG, X, NewInfoX),
   InfoY = hipe_vectors:get(IG, Y),
   OldYMoves = InfoY#node.moves,
   NewInfoY = InfoY#node{moves=[X|OldYMoves]},
   hipe_vectors:set(IG0, Y, NewInfoY).

add_edge(X, X, IG) ->
   IG;
add_edge(X, Y, IG) ->
   add_arc(X, Y, add_arc(Y, X, IG)).


add_arc(X, Y, IG) ->
   case hipe_sparc_registers:is_precolored(X) of
      true ->
	 IG;
      false ->
	 Info = hipe_vectors:get(IG, X),
	 Old = Info#node.adj,
	 New = Info#node{adj=[Y|Old]},
	 hipe_vectors:set(IG,X,New)
   end.


normalize_ig(IG) ->
   Size = hipe_vectors:vsize(IG),
   normalize_ig(Size, IG).

normalize_ig(0, IG) ->
   IG;
normalize_ig(I, IG) ->
   Info = hipe_vectors:get(IG, I),
   N = ordsets:from_list(Info#node.adj),
   M = ordsets:from_list(Info#node.moves),
   Deg = case hipe_sparc_registers:is_precolored(I) of
	    true ->
	       2345433;  % Approximation of infinity ;)
	    false ->
	       length(N)
	 end,
   NewInfo = Info#node{adj=N, degree=Deg, moves=M},
   NewIG = hipe_vectors:set(IG, I, NewInfo),
   normalize_ig(I-1, NewIG).


degree(X, IG) ->
   Info = hipe_vectors:get(IG, X),
   Info#node.degree.

set_degree(IG, X, D) ->
   Info = hipe_vectors:get(IG, X),
   NewInfo = Info#node{degree=D},
   NewIG = hipe_vectors:set(IG, X, NewInfo).


moves(X, IG) ->
   Info = hipe_vectors:get(IG, X),
   Info#node.moves.

adj(X, IG) ->
   Info = hipe_vectors:get(IG, X),
   Info#node.adj.


list_ig(IG) ->
   hipe_vectors:list(IG).


%
%
%

combine(IG, X, Y, AdjX, AdjY, K) ->
   AdjXY = ordsets:union(AdjX, AdjY),
   IG0 = update_edges(IG, X, Y, AdjY),
   MovesY = moves(Y, IG0),
   IG1 = update_moves(IG0, X, Y, MovesY),
   InfoX = hipe_vectors:get(IG1, X),
   MovesX = ordsets:union(InfoX#node.moves, MovesY),
   DegX = calc_degree(AdjXY, IG, K),
   hipe_vectors:set(IG1, X, InfoX#node{adj=AdjXY, moves=ordsets:del_element(X, MovesX)}).


calc_degree([], IG, K) ->
   0;
calc_degree([N|Ns], IG, K) ->
   Info = hipe_vectors:get(IG, N),
   Deg = Info#node.degree,
   Moves = Info#node.moves,
   if Deg < K,  Moves =:= []->
	 calc_degree(Ns, IG, K);
      true ->
	 1 + calc_degree(Ns, IG, K)
   end.


update_edges(IG, X, Y, []) ->
   IG;
update_edges(IG, X, Y, [A|As]) ->
   InfoA = hipe_vectors:get(IG, A),
   Nb = InfoA#node.adj,
   Nb0 = ordsets:del_element(Y, Nb),
   Nb1 = ordsets:add_element(X, Nb0),
   NewInfoA = InfoA#node{adj=Nb1},
   update_edges(hipe_vectors:set(IG,A,NewInfoA), X, Y, As).


update_moves(IG, X, Y, []) ->
   IG;
update_moves(IG, X, Y, [M|Ms]) ->
   InfoM = hipe_vectors:get(IG, M),
   Moves = InfoM#node.moves,
   Moves0 = ordsets:del_element(Y, Moves),
   Moves1 = ordsets:add_element(X, Moves0),
   NewInfoM = InfoM#node{moves=Moves1},
   update_moves(hipe_vectors:set(IG,M,NewInfoM), X, Y, Ms).



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% The coloring datatype:

none_colored(NumNodes) ->
   hipe_vectors:init(NumNodes,uncolored).

color_of(X,Cols) ->
   hipe_vectors:get(Cols,X).

set_color(X,R,Cols) ->
   hipe_vectors:set(Cols,X,{color,R}).

list_coloring(Cols) ->
   hipe_vectors:list(Cols).



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Check that all arcs in IG are bidirectional + degree is correct
%

check_ig(IG) ->
   check_ig(list_ig(IG),IG).

check_ig([],IG) -> 
   ok;
check_ig([{N,Info}|Xs],IG) ->
   Ns = adj(Info),
   NumNs = length(Ns),
   D = degree(Info),
   if
      D == NumNs ->
	 ok;
      true ->
	 warning('node ~p has degree ~p but ~p adj~n',[N,D,NumNs])
   end,
   check_adj(N,Ns,IG),
   check_ig(Xs,IG).

check_adj(N,[],IG) -> 
   ok;
check_adj(N,[M|Ms],IG) ->
   Ns = adj(M,IG),
   case member(N,Ns) of
      true ->
	 ok;
      true ->
	 warning('node ~p should have ~p as neighbor (has ~p)~n',[M,N,Ns])
   end,
   check_adj(N,Ms,IG).



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Check that the coloring is correct (if the IG is correct):
%
% NOTE: does NOTNOTNOT use same representation of coloring as
%  the standard allocator!
%

check_coloring(Coloring, IG) ->
   check_cols(list_ig(IG), init_coloring(Coloring)).

init_coloring(Xs) ->
   hash:init(Xs).

check_color_of(X, Cols) ->
   case hash:lookup(X, Cols) of
      not_found ->
	 uncolored;
      {found,C} ->
	 C
   end.

check_cols([],Cols) ->
   ok;
check_cols([{X,Info}|Xs],Cols) ->
   Cs = [{N,check_color_of(N,Cols)} || N <- adj(Info)],
   C = check_color_of(X,Cols),
   case valid_coloring(X,C,Cs) of
      yes ->
	 check_cols(Xs,Cols);
      {no,Invalids} ->
	 warning(check_cols, "node ~w has same color (~w) as ~w~n",
		 [X,C,Invalids]),
	 check_cols(Xs,Cols)
   end.

valid_coloring(X,C,[]) ->
   yes;
valid_coloring(X,C,[{Y,C}|Ys]) ->
   case valid_coloring(X,C,Ys) of
      yes -> {no,[Y]};
      {no,Zs} -> {no,[Y|Zs]}
   end;
valid_coloring(X,C,[_|Ys]) ->
   valid_coloring(X,C,Ys).



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% *** ERROR HANDLING ***
%

warning(Fun,Msg) ->
   io:format('WARNING (~p): ~p~n',[Fun,Msg]).

warning(Fun,Msg,Xs) ->
   io:format('WARNING (~p):',[Fun]),
   io:format(Msg,Xs),
   io:format('~n',[]).

error(Fun,Msg) ->
   io:format('ERROR (~p): ~p~n',[Fun,Msg]),
   exit({error,Fun,Msg}).



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% *** INTERFACES TO OTHER MODULES ***
%



reg_names([]) ->
   [];
reg_names([R|Rs]) ->
   [hipe_sparc:reg_nr(R) | reg_names(Rs)].


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Precoloring: use this version when a proper implementation of
%  physical_name(X) is available!
%

precolor(Xs, Cols) ->
   precolor0(Xs, Cols).

precolor0([], Cols) ->
   {[], Cols};
precolor0([R|Rs], Cols) ->
   {Cs, Cols1} = precolor0(Rs, Cols),
   {[{R,{reg,physical_name(R)}}|Cs], set_color(R,physical_name(R),Cols1)}.


physical_name(X) ->
   hipe_sparc_registers:physical_name(X).


print_ig(IG) ->
   lists:foreach(
     fun({N, II}) -> io:format("*~w*(~w): ~w,~w~n", 
			       [N,degree(II),moves(II),adj(II)]) 
     end,
     lists:keysort(1, list_ig(IG))).





