%% Copyright (c) 2000 by Erik Johansson.  All Rights Reserved 
%% =====================================================================
%%  Filename : 	hipe_ls_regalloc.erl
%%  Module   :	hipe_ls_regalloc
%%  Purpose  :  Perform a register allocation based on the 
%%              "linear-scan algoritm".
%%  Notes    :  * This is an implementation of 
%%                "Linear Scan Register Allocation" by 
%%                Massimiliano Poletto & Vivek Sarkar described in
%%                ACM TOPLAS Vol 21, No 5, September 1999.
%%
%%              * This implementation is sparc-dependent but all 
%%                interface functions to sparc specifics are gathered at
%%                the end of the file. 
%%                (Still waiting for a modular module system for Erlang.)
%%
%%              
%%  History  :	* 2000-04-07 Erik Johansson (happi@csd.uu.se): Created.
%% CVS:
%%    $Author: richardc $
%%    $Date: 2001/03/26 18:37:08 $
%%    $Revision: 1.1.1.1 $
%% =====================================================================
%% Exported functions (short description):
%%   regalloc(CFG,PhysRegs,Entrypoints) -> {Coloring, NumberOfSpills}
%%    Takes a (SPARC) CFG and returns a coloring of all used registers.  
%%    PhysRegs should be a list of availabel physical registers.
%%    Entrypoints should be a list of names of Basic Blocks that have
%%     external entrypoints.
%%
%%    The Coloring will be in the form of the "allocation datastructure"
%%     described below, that is, a list of tuples on the form
%%      {Name, {reg, PhysicalRegister}} or
%%      {Name, {spill, SpillIndex}}
%%    The NumberOfSpills is either 0 indicating no spill or the 
%%     SpillIndex of the last spilled register.
%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

-module(hipe_ls_regalloc).
-export([regalloc/4]).

%-define(DEBUG(__What,__X),
%	case catch lists:member(__What,get(hipe_debug)) of
%	    true -> __X; _ -> ok end).
%-define(DEBUG(What,X),X).
-define(DEBUG(What,X),no_debug).

-define(count_regs(X),X).
 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%
%% regalloc(CFG,PhysRegs,Entrypoints) 
%%   Calculates an allocation of registers using a linear_scan algorithm.
%%   There are three steps in the algorithm:
%%    1. Calculate live-ranges for all registers.
%%    2. Calculate live-intervalls for each register.
%%       The live intervall consits of a start position and a endpostion
%%       these are the first defintion and last use of the register 
%%       given as instruction numbers in a breth-first traversal of the
%%       controll-flow-graph.
%%    3. Do a linear scan allocation over the live intervalls.
%%
%%-  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -
regalloc(CFG,PhysRegs,Entrypoints, Options) ->
  ?DEBUG(time,io:format("LinearScan: ~w\n",[erlang:statistics(runtime)])),
  %%     Step 1: Calculate liveness (Call external implementation.)
  Liveness = liveness(CFG),
  ?DEBUG(time,io:format("liveness (done)~w\n",[erlang:statistics(runtime)])),
  USIntervals = calculate_intervals(CFG,Liveness,Entrypoints, Options),
  ?DEBUG(time,io:format("intervalls (done) ~w\n",[erlang:statistics(runtime)])),
  Intervals = sort_on_start(USIntervals),
  ?DEBUG(time,io:format("sort intervalls (done) ~w\n",[erlang:statistics(runtime)])),
  ?DEBUG(time,io:format("No intervalls: ~w\n",[length(Intervals)])),
  ?count_regs(put(noregs,get(noregs)+length(Intervals))),
  ?DEBUG(time,io:format("count intervalls (done) ~w\n",[erlang:statistics(runtime)])),
  Allocation = allocate(Intervals,PhysRegs),
  ?DEBUG(time,io:format("allocation (done) ~w\n",[erlang:statistics(runtime)])),
  Allocation.
%%^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%                                                                    %%
%%        Step 2: Calculate live-intervalls for each register.        %%
%%                                                                    %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%-  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -
%% calculate_intervals(CFG,Liveness,Entrypoints)
%%  CFG: The Control-Flow Graph.
%%  Liveness: A map of live-in and live-out sets for each Basic-Block.
%%  Entrypoints: A set of BB names that have external entrypoints.
%%
calculate_intervals(CFG,Liveness,Entrypoints, Options) ->
  %% Add start point for the argument registers.
  {_,_,Arity} = hipe_sparc_cfg:function(CFG),
  Args = arg_vars(Arity),
  Interval = add_livepoint(Args, 0, empty_interval()),
  Worklist =
    case property_lists:get_value(ls_order, Options) of
      reversepostorder ->
	hipe_sparc_cfg:reverse_postorder(CFG);
      breadth ->
	hipe_sparc_cfg:breadthorder(CFG);
      postorder ->
	hipe_sparc_cfg:postorder(CFG);
      prediction ->
	hipe_sparc_cfg:predictionorder(CFG);
      random ->
	hipe_sparc_cfg:labels(CFG);
      _ ->
	hipe_sparc_cfg:reverse_postorder(CFG)
    end,

  ?DEBUG(ls,io:format("No BBs ~w\n",[length(Worklist)])),
  intervals(Worklist, Interval, 1, CFG, Liveness, succ_map(CFG)).

%%-  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -
%% intervals(WorkList, Intervals, InstructionNr,
%%           CFG, Liveness, SuccMap)
%%  WorkList: List of BB-names to handle.
%%  Intervals: Intervals seen so far (sorted on register names).
%%  InstructionNr: The number of examined insturctions.
%%  CFG: The Control-Flow Graph.
%%  Liveness: A map of live-in and live-out sets for each Basic-Block.
%%  SuccMap: A map of successors for each BB name.
%%-  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -
intervals([L|ToDO],Intervals,InstructionNr,CFG,Liveness,SuccMap) ->
  %% Add all variables that are live at the entry of this block
  %% to the interval data structure.
  LiveIn = livein(Liveness,L),
  Intervals2 = add_livepoint(LiveIn,InstructionNr,Intervals),
  LiveOut = liveout(Liveness,L),
  Unchanged = ordsets:intersection(LiveIn, LiveOut),
  %% Traverse this block instruction by instruction and add all
  %% uses and defines to the intervals.

  Code = hipe_bb:code(bb(CFG,L)),
  {Intervals3, NewINr} = traverse_block(Code, InstructionNr,
					Intervals2, Unchanged),
  
  %% Add end points for the registers that are in the live-out set.
  Intervals4 = add_livepoint(LiveOut, NewINr, Intervals3),
  
  intervals(ToDO, Intervals4, NewINr, CFG, Liveness, SuccMap);
intervals([],Intervals,_,_,_,_) -> 
  %% Return the calculated intervals
  Intervals.

%%-  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -
%% traverse_block(Code, InstructionNo, Intervals, Unchanged) 
%%  Examine each instruction in the Code:
%%   For each temporary T used or defined by instruction number N:
%%    extend the interval of T to include N.
%%-  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -
traverse_block([Instruction|Is],InstrNo,Intervals, Unchanged) ->
  %% Get used temps.
  UsesSet = ordsets:from_list(uses(Instruction)),

  %% Get defined temps.
  DefsSet = ordsets:from_list(defines(Instruction)),

  %% Only consider those temps that starts or ends their lifetime 
  %%  within the basic block (that is remove all Unchanged temps).
  Update = ordsets:subtract(ordsets:union(UsesSet, DefsSet), Unchanged),

  %% Extend the intervals for these temporaries to include InstrNo.
  Intervals1 = add_livepoint(ordsets:to_list(Update), InstrNo, Intervals),

  %% Handle the next instruction.
  traverse_block(Is,InstrNo+1,Intervals1, Unchanged);

traverse_block([], InstrNo, Intervals, _) -> 
  %% Return the new intervals and the number of the next instruction.
  {Intervals,InstrNo}.
%%^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%                                                                    %%
%%    Step 3. Do a linear scan allocation over the live intervalls.   %%
%%                                                                    %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%
%% allocate(Intervals, PhysicalRegisters)
%%
%% This function performs the linear scan algorithm.
%%  Intervals contains the start and stop position of each register,
%%            sorted on increasing startpositions
%%  PhysicalRegisters is a list of available Physical registers to use.
%%
%%-  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -
allocate(Intervals, PhysRegs) ->
  ActiveRegisters =[],
  AllocatedRegisters = empty_allocation(),
  SpillIndex = 0,
  allocate(Intervals, PhysRegs, ActiveRegisters,
	   AllocatedRegisters, SpillIndex).
%%-  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -
%% allocate(Intervalls, Free, Active, Allocated, SpillIndex) 
%%  Iterates of each register interval.
%%   Intervalls: The list of register intervalls.
%%   Free: Currently available physical registers.
%%   Active: Currently used physical registers (sorted on increasing 
%%            intervall enpoints)
%%   Allocated: The mapping of register names to physical registers or
%%              to spill positions.
%%   SpillIndex: The number of spilled registers. 
%%-  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -
allocate([RegInt|RIS], Free, Active, Alloc, SpillIndex) ->
  %% io:format("Alloc intervall: ~w, Free ~w\n",[RegInt, Free]),
  %% Remove from the active list those registers who's intervals 
  %% ends before the start of the current interval.
  {NewActive, NewFree} = expire_old_intervals(Active, startpoint(RegInt), Free),
  
  %% Get the name of the temp in the current interval.
  Temp = reg(RegInt), 
  case is_precolored(Temp) of
    true -> 
      %% This is a precolored register we don't need to find a color
      %% Get the physical name of the register.
      PhysName = physical_name(Temp), 
      %% Bind it to the precolored name.
      NewAlloc = alloc(Temp, PhysName, Alloc), 
      case is_global(Temp) of
	true -> 
	  %% this is a global precolored register 
	  allocate(RIS, NewFree, NewActive,
		   NewAlloc, SpillIndex);
	false ->
	  case is_free(PhysName, NewFree) of
	    true ->
	      allocate(RIS, NewFree -- [PhysName], 
		       add_active(endpoint(RegInt), PhysName, Temp, NewActive),
		       NewAlloc,
		       SpillIndex);
	    false ->
	      %% Some other temp has taken this precolored register,
	      %% throw it out.

	      {OtherActive, NewActive2} = 
		deactivate(PhysName, NewActive),
	      OtherTemp = active_name(OtherActive),
	      OtherEnd = active_endpoint(OtherActive),
	      OtherInt = mk_interval(OtherTemp,0,OtherEnd),
	      NewSpillIndex = SpillIndex+1,
	      {NewAlloc2, NewActive3} = 
		spill(OtherTemp, OtherEnd, NewActive2, NewAlloc, NewSpillIndex),
	      allocate(RIS, 
		       NewFree, 
		       add_active(endpoint(RegInt), PhysName, Temp, NewActive3),
		       NewAlloc2, NewSpillIndex)
	  
	  end
	    
      end;
    false -> 
      %% This is not a precolored register.
      case NewFree of 
	[] -> 
	  %% No physical registers available, we have to spill.
	  NewSpillIndex = SpillIndex+1,
	  {NewAlloc, NewActive2} = 
	    spill(Temp, endpoint(RegInt), Active, Alloc, NewSpillIndex),
	  %% io:format("Spilled ~w\n",[NewAlloc]),
	  allocate(RIS, NewFree, NewActive2, NewAlloc, NewSpillIndex);

	[FreeReg | Regs] -> 
	  %% The register FreeReg is available, let's use it.
	  allocate(RIS,Regs,
		   add_active(endpoint(RegInt), FreeReg, Temp, NewActive),
		   alloc(Temp, FreeReg, Alloc),
		   SpillIndex)
      end
  end;
allocate([],_,_,Alloc,SpillIndex) -> 
  %% No more register intervals to handle
  %%  return the result.
  {Alloc, SpillIndex}.
%%^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%
%% expire_old_intervals(ActiveRegisters, CurrentPos, FreeRegisters) 
%%   Remove all registers that have live-ranges that ends before the
%%   current position from the active list and put them into the free
%%   list instead.
%%
%% ---------------------------------------------------------------------
expire_old_intervals([Active|Actives], CurrentPos, Free) ->
  %% Does the live-range of the first active register end before 
  %% the current position?
  %% We don't free registers that end at the current position,
  %%  since a multimove can decide to do the moves in another order...
  case active_endpoint(Active) < CurrentPos of
    true -> %% Yes -> Then we can free that register.

		
      expire_old_intervals(Actives, CurrentPos,
			   %% Add the register to the free pool.
			   [active_reg(Active)|Free]);
			   
			   %% Here we could try appending the
			   %% register to get a more widespread
			   %% use of registers.
			   %% Free ++ [active_reg(Active)]);
                           %% At the moment this does not seem to
                           %%  improve performance at at,
                           %%  on the other hand, the cost is very low.


    false -> 
      %% No -> Then we cannot free any more registers.
      %%       (Since they are sorted on endpoints...)    
      {[Active|Actives],Free}
  end;
expire_old_intervals([],_,Free) ->
  {[],Free}.
%%^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

deactivate(Reg, [Active|Actives]) ->
  case Reg =:= active_reg(Active) of
    true ->
      {Active, Actives};
    false ->
      {TheActive, NewActives} =
	deactivate(Reg, Actives),
      {TheActive, [Active|NewActives]}
  end;
deactivate(_,[]) -> {no,[]}.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%
%% spill(CurrentReg, CurrentEndpoint, Active, Alloc, SpillIndex)
%%   Find the register with the longest live range and spill it to memory.
%%
%% ---------------------------------------------------------------------
spill(CurrentReg, CurrentEndpoint, Active, Alloc, SpillIndex) ->
  %%  io:format("spilling",[]),

  %% Find a spill candidate (one of the active): 
  %%  The register with the longest live-range.
  {NewActive, SpillCandidate} = butlast_last(Active),

  SpillEndpoint = active_endpoint(SpillCandidate) ,
  SpillName = active_name(SpillCandidate),
  SpillPhysName = active_reg(SpillCandidate), 

  case SpillEndpoint > CurrentEndpoint of
    true -> 
      %% There is an already allocated register that has
      %% a longer live-range than the current register.
      case is_precolored(SpillName) of 
	true ->
	  %% Cannot spill a precolored register
	  {NewAlloc, NewActive2} = 
	    spill(CurrentReg, CurrentEndpoint, NewActive, Alloc, SpillIndex),
	  {NewAlloc, 
	   add_active(SpillEndpoint, SpillPhysName, SpillName,
		      NewActive2)};
	false ->
	  %% It is not precolored...

	  %% Allocate SpillCandidate to spill-slot SpillIndex
	  SpillAlloc = 
	    spillalloc(active_name(SpillCandidate), SpillIndex, 
		       Alloc),

	  %% Allocated the current register to the physical register
	  %% used by the spill candidate.
	  NewAlloc = alloc(CurrentReg, SpillPhysName, SpillAlloc),
	  
	  %% Add the current register to the active registers
	  NewActive2 = 
	    add_active(CurrentEndpoint, SpillPhysName, CurrentReg, NewActive),
	  
	  {NewAlloc, NewActive2}
      end;
	
    false -> 
      %% The current register has the longest live-range.

      case is_precolored(CurrentReg) of 
	true ->
	  %% Cannot spill a precolored register
	  {NewAlloc, NewActive2} = 
	    spill(SpillName, SpillEndpoint, NewActive, Alloc, SpillIndex),
	  NewActive3 = 
	    add_active(CurrentEndpoint, SpillPhysName, CurrentReg, NewActive2),
	  {NewAlloc, NewActive3};
	false ->
	  %% It is not precolored...
	  %% Allocate the current register to spill-slot SpillIndex
	  {spillalloc(CurrentReg, SpillIndex, Alloc), Active}
      end
  end.
%%^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%                                                                    %%
%%                   D A T A   S T R U C T U R E S                    %%
%%                                &                                   %%
%%               A U X I L I A R Y   F U N C T I O N S                %%
%%                                                                    %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%




%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%
%% The "allocation datastructure"
%%
%% This is an order list of register names paired with their allocations.
%%  {Name, Allocation}
%% The allocation is either {reg, physical register} or
%%                          {spill, spill index}
%%
%% ---------------------------------------------------------------------
empty_allocation() -> [].

alloc(Name,Reg,[{Name,_}|A]) ->
  [{Name,{reg,Reg}}|A];
alloc(Name,Reg,[{Name2,Binding}|Bindings]) when Name > Name2 ->
  [{Name2,Binding}|alloc(Name,Reg,Bindings)];
alloc(Name,Reg,Bindings) ->
  [{Name,{reg,Reg}}|Bindings].

spillalloc(Name,N,[{Name,_}|A]) ->
  [{Name,{spill,N}}|A];
spillalloc(Name,N,[{Name2,Binding}|Bindings]) when Name > Name2 ->
  [{Name2,Binding}|spillalloc(Name,N,Bindings)];
spillalloc(Name,N,Bindings) ->
  [{Name,{spill,N}}|Bindings].
%%^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%
%%
butlast_last([X]) ->
   {[],X};
butlast_last([X|Y]) ->
  {L,Last} = butlast_last(Y),
  {[X|L],Last}.

%butlast([X]) ->
%   [];
% butlast([X|Y]) ->
%  [X|butlast(Y)];
%butlast([])->
%  [].

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% 
%%  The active datastructure.
%%   Keeps tracks of currently active (allocated) physical registers.
%%   It is sorted on end points in the intervals
%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
add_active(Endpoint, PhysReg, RegName, [{P1,R1,O1}|Active]) when P1 < Endpoint ->
  [{P1,R1,O1}|add_active(Endpoint, PhysReg, RegName, Active)];
add_active(Endpoint, PhysReg, RegName, Active) ->
  [{Endpoint, PhysReg, RegName}|Active].

active_reg({_,PhysReg,_}) ->
  PhysReg.
active_endpoint({EndPoint,_,_}) ->
  EndPoint.
active_name({_,_,RegName})->
  RegName.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% The Interval data structure.
%%
%%
%%-  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -
empty_interval() ->
  gb_trees:empty().

sort_on_start(I)->
 lists:keysort(2,gb_trees:to_list(I)).

mk_interval(Name, Start, End) ->
  {Name, { Start, End}}.

endpoint({R,{S,Endpoint}}) ->
  Endpoint.
startpoint({R,{Startpoint,E}}) ->
  Startpoint.
reg({RegName,{S,E}}) ->
  RegName.

%linarize(I) ->
%  linarize(I,[]).

%linarize({R,Beg,End,Lt,Rt},Acc) ->
%  linarize(Lt,[{R,Beg,End}|linarize(Rt,Acc)]);
%linarize([],Acc) -> Acc.

add_livepoint([Temp|Temps],Pos,Intervals) ->

   
  %% Extend the old interval...
  NewInterval =
    case gb_trees:lookup(Temp, Intervals) of
      %% This temp has an old interval...
      {value, Value} ->
	%% ... extend it.
	extend_interval(Pos, Value);

      %% This is the first time we see this temp...
      none ->
	%% ... create a new interval
	{Pos, Pos}
    end,

  %% Add or update the extended interval.
  Intervals2 = gb_trees:enter(Temp, NewInterval, Intervals),

  %% Add the rest of teh temporaries.
  add_livepoint(Temps, Pos, Intervals2);

add_livepoint([], _, I) ->
  %% No more to add return the interval.
  I.

extend_interval(Pos, {Beginning, End}) ->
  %% If this position occures after the end
  %%  of the interval, then extend the end to
  %%  this position.
  NewEnd = 
    if Pos > End -> Pos;
       true      -> End
    end,

  %% If this position occures before the beginning
  %%  of the interval, then extend the beginning to
  %%  this position.

  NewBeginning = 
    if Pos < Beginning -> Pos;
       true            -> Beginning
    end,

  {NewBeginning, NewEnd}. 


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% The Freel data structure.
%%
%%
%%-  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -

is_free(R, [R|_]) ->
  true;
is_free(R, [_|Rs]) ->
  is_free(R, Rs);
is_free(_, [] ) ->
  false.

%%^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%
%% Interface to external functions.
%% XXX: Make this modular somehow...
%% 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
succ_map(CFG) ->
  hipe_sparc_cfg:succ_map(CFG).

liveness(CFG) ->
   hipe_sparc_liveness:analyze(CFG).

bb(CFG,L) ->
  hipe_sparc_cfg:bb(CFG,L).

livein(Liveness,L) ->
  regnames(hipe_sparc_liveness:livein(Liveness,L)).


liveout(Liveness,L)->
  regnames(hipe_sparc_liveness:liveout(Liveness,L)).

%succ(SuccMap, L)->
%  hipe_sparc_cfg:succ(SuccMap, L).

uses(I)->
  regnames(hipe_sparc:uses(I)).

defines(I) ->
  regnames(hipe_sparc:defines(I)).

is_precolored(R) ->
  hipe_sparc_registers:is_precolored(R).

is_global(R) ->
  lists:member(R,hipe_sparc_registers:all_precolored() -- all_args()).

all_args() ->
  [hipe_sparc_registers:arg(0),
   hipe_sparc_registers:arg(1),
   hipe_sparc_registers:arg(2),
   hipe_sparc_registers:arg(3),
   hipe_sparc_registers:arg(4)].

physical_name(R) ->
  hipe_sparc_registers:physical_name(R).

regnames([R|Rs]) ->
  [hipe_sparc:reg_nr(R)|regnames(Rs)];
regnames([]) -> [].



arg_var(X) ->
   hipe_sparc_registers:arg(X).

arg_vars(N, Acc) when N >= 0 ->
  arg_vars(N-1, [arg_var(N)|Acc]);
arg_vars(_, Acc) -> Acc.

arg_vars(N) ->
  case N >= hipe_sparc_registers:register_args() of
    false ->
      arg_vars(N-1,[]);
    true ->
      arg_vars(hipe_sparc_registers:register_args()-1,[])
  end.
