%%% $Id: hipe_x86_frame.erl,v 1.17 2001/06/13 14:45:39 mikpe Exp $
%%% x86 stack frame handling
%%%
%%% - map non-register temps to stack slots
%%% - add explicit stack management code to prologue and epilogue,
%%%   and at calls and tailcalls
%%%
%%% XXX:
%%% - It is getting a bit messy to do the on-the-fly mapping from
%%%   virtual FP offsets to actual SP offsets. Move FP elimination
%%%   to a second pass?

-module(hipe_x86_frame).
-export([frame/1]).
-include("hipe_x86.hrl").

frame(Defun) ->
    Formals = hipe_x86:defun_formals(Defun),
    Temps = all_temps(hipe_x86:defun_code(Defun), Formals),
    CFG0 = hipe_x86_cfg:init(Defun),
    Liveness = hipe_x86_liveness:analyse(CFG0),
    init_exnlabs(),
    CFG1 = do_body(CFG0, Liveness, Formals, Temps),
    CFG2 = hipe_x86_cfg:var_range_update(CFG1, []),
    CFG3 = finalise_exnlabs(CFG2),
    hipe_x86_cfg:linearise(CFG3).

do_body(CFG0, Liveness, Formals, Temps) ->
    Context = mk_context(Liveness, Formals, Temps),
    CFG1 = do_blocks(CFG0, Context),
    do_prologue(CFG1, Context).

do_blocks(CFG, Context) ->
    Labels = hipe_x86_cfg:labels(CFG),
    do_blocks(Labels, CFG, Context).

do_blocks([Label|Labels], CFG, Context) ->
    Liveness = context_liveness(Context),
    LiveOut = hipe_x86_liveness:liveout(Liveness, Label),
    Block = hipe_x86_cfg:bb(CFG, Label),
    Code = hipe_bb:code(Block),
    NewCode = do_block(Code, LiveOut, Context),
    NewBlock = hipe_bb:code_update(Block, NewCode),
    NewCFG = hipe_x86_cfg:bb_update(CFG, Label, NewBlock),
    do_blocks(Labels, NewCFG, Context);
do_blocks([], CFG, _) ->
    CFG.

do_block([I|Insns], LiveOut, Context) ->
    NewIs = do_insn(I, LiveOut, Context),
    NewIs ++ do_block(Insns, LiveOut, Context);
do_block([], _, _) ->
    [].

do_insn(I, LiveOut, Context) ->	% Insn -> [Insn]
    case I of
	#alu{} ->
	    [do_alu(I, Context)];
	#cmp{} ->
	    [do_cmp(I, Context)];
	#move{} ->
	    [do_move(I, Context)];
	#pseudo_call{} ->
	    do_pseudo_call(I, LiveOut, Context);
	#pseudo_return{} ->
	    do_pseudo_return(I, Context);
	#pseudo_tailcall{} ->
	    do_pseudo_tailcall(I, Context);
	_ ->	% comment, pseudo_jcc, jmp, label
	    [I]
    end.

%%%
%%% Convert any pseudo-temp operand in a binary instruction (alu, cmp, move)
%%% to an explicit x86_mem operand.
%%%

do_alu(I, Context) ->
    #alu{src=Src0,dst=Dst0} = I,
    Src = conv_opnd(Src0, Context),
    Dst = conv_opnd(Dst0, Context),
    I#alu{src=Src,dst=Dst}.

do_cmp(I, Context) ->
    #cmp{src=Src0,dst=Dst0} = I,
    Src = conv_opnd(Src0, Context),
    Dst = conv_opnd(Dst0, Context),
    I#cmp{src=Src,dst=Dst}.

do_move(I, Context) ->
    #move{src=Src0,dst=Dst0} = I,
    Src = conv_opnd(Src0, Context),
    Dst = conv_opnd(Dst0, Context),
    I#move{src=Src,dst=Dst}.

conv_opnd(Opnd, Context) ->
    conv_opnd(Opnd, context_framesize(Context), Context).

conv_opnd(Opnd, FPoff, Context) ->
    case opnd_is_pseudo(Opnd) of
	false ->
	    Opnd;
	true ->
	    conv_pseudo(Opnd, FPoff, Context)
    end.

conv_pseudo(Temp, FPoff, Context) ->
    Off = FPoff + context_offset(Context, Temp),
    hipe_x86:mk_mem(mk_esp(), hipe_x86:mk_imm(Off), hipe_x86:temp_type(Temp)).

%%%
%%% Return - deallocate frame and emit 'ret $N' insn.
%%%

do_pseudo_return(I, Context) ->
    adjust_esp(context_framesize(Context),
	       [hipe_x86:mk_ret(4*context_arity(Context))]).

adjust_esp(N, Rest) ->
    if N =:= 0 ->
	    Rest;
       true ->
	    [hipe_x86:mk_alu('add', hipe_x86:mk_imm(N), mk_esp()) | Rest]
    end.

%%%
%%% Recursive calls.
%%% 1. if has exn handler, push catch word [until stack map is impl]
%%% 2. push actuals left-to-right
%%% 3. write zeros to dead stack slots [until stack map is impl]
%%% 4. call
%%% 5. if has exn handler, pop catch word [until stack map is impl]
%%% 6. jump to continuation label

do_pseudo_call(I, LiveOut, Context) ->
    JmpCode = [hipe_x86:mk_jmp_label(hipe_x86:pseudo_call_contlab(I))],
    ExnLab = hipe_x86:pseudo_call_exnlab(I),
    ReturnCode =
	case ExnLab of
	    [] ->
		JmpCode;
	    _ ->
		[hipe_x86:mk_alu('add', hipe_x86:mk_imm(4), mk_esp()) |
		 JmpCode]
	end,
    OrigArgs = hipe_x86:pseudo_call_args(I),
    RealArgs =
	case ExnLab of
	    [] ->
		OrigArgs;
	    _ ->
		add_exnlab(ExnLab),
		[hipe_x86:mk_imm_from_addr(ExnLab, 'catch') | OrigArgs]
	end,
    FPoff0 = context_framesize(Context),
    FPoff1 = FPoff0 + 4*length(RealArgs),
    Fun0 = hipe_x86:pseudo_call_fun(I),
    Fun1 = conv_opnd(Fun0, FPoff1, Context),
    CallCode = [hipe_x86:mk_call(Fun1) | ReturnCode],
    ZapCode = do_zap_code(LiveOut, FPoff1, Context, CallCode, Fun0),
    do_call_args(RealArgs, FPoff0, Context, ZapCode).

do_call_args([Arg|Args], FPoff, Context, Tail) ->
    [hipe_x86:mk_push(conv_opnd(Arg, FPoff, Context)) |
     do_call_args(Args, FPoff+4, Context, Tail)];
do_call_args([], FPoff, Context, Tail) ->
    %% +4 for our RA and +4 for callee's RA should it need to call inc_stack
    context_need_stack(Context, FPoff + 8),
    Tail.

do_zap_code(LiveOut, FPoff, Context, Tail, Fun) ->
    Temps = context_temps(Context),
    Dead = tset_del_list(tset_del_list(Temps, LiveOut),
			 hipe_x86_defuse:src_use(Fun)),
    case Dead of
	[] ->
	    Tail;
	_ ->
	    %% "movl $0,off(%esp)" requires an imm32 zero so we clear %eax
	    %% and "movl %eax,off(%esp)" -- saves a lot of object code
	    EAX = mk_eax('untagged'),
	    I = hipe_x86:mk_alu('xor', EAX, EAX),
	    [I | zap_each(Dead, FPoff, Context, EAX, Tail)]
    end.

zap_each([Temp|Temps], FPoff, Context, Zero, Tail) ->
    [zap_temp(Temp, FPoff, Context, Zero) |
     zap_each(Temps, FPoff, Context, Zero, Tail)];
zap_each([], _, _, _, Tail) ->
    Tail.

zap_temp(Temp, FPoff, Context, Zero) ->
    hipe_x86:mk_move(Zero, conv_pseudo(Temp, FPoff, Context)).

%%%
%%% Tailcalls.
%%%

do_pseudo_tailcall(I, Context) ->
    FPoff0 = context_framesize(Context),
    Arity = context_arity(Context),
    Args = hipe_x86:pseudo_tailcall_args(I) ++ [context_ra(Context)],
    Fun0 = hipe_x86:pseudo_tailcall_fun(I),
    {Insns, FPoff1, Fun1} = do_tailcall_args(Args, Context, Fun0),
    context_need_stack(Context, FPoff1),
    FPoff2 = FPoff1 + 4+4*Arity - 4*length(Args),
    context_need_stack(Context, FPoff2 + 4),	% +4 for callee's inc_stack RA
    I2 = hipe_x86:mk_jmp_fun(Fun1),
    Insns ++ adjust_esp(FPoff2, [I2]).

do_tailcall_args(Args, Context, Fun0) ->
    FPoff0 = context_framesize(Context),
    Arity = context_arity(Context),
    FrameTop = 4 + 4*Arity,
    DangerOff = FrameTop - 4*length(Args),
    Moves = mk_moves(Args, FrameTop, []),
    {Stores, Simple, Conflict} =
	split_moves(Moves, Context, DangerOff, [], [], []),
    {GrowStack, FPoff1} =
	if DangerOff < -FPoff0 ->
		%% XXX: We should ideally have set up a larger initial stack
		%% frame to avoid this. Would have to take tailcalls into
		%% account before computing Context, and "fake" unused temps
		%% in these stack slots to ensure correct behaviour at
		%% recursive calls. Later.
		NextFPoff = -DangerOff,
		Increment = NextFPoff - FPoff0,
		io:format(standard_io,"do_tailcall_args: DangerOff ~w FPoff0 ~w; growing stack with ~w\n", [DangerOff, FPoff0, Increment]),
		{[hipe_x86:mk_alu('sub', hipe_x86:mk_imm(Increment), mk_esp())],
		 NextFPoff};
	   true ->
		{[], FPoff0}
	end,
    {Pushes, MoreSimple, FPoff2} = split_conflict(Conflict, FPoff1, [], []),
    {FixFun1, Fun1} =
	case opnd_is_pseudo(Fun0) of
	    false ->
		{[], Fun0};
	    true ->
		%% use edx since simple_moves clobbers eax
		EDX = mk_edx(),
		MEM = conv_pseudo(Fun0, FPoff2, Context),
		{[hipe_x86:mk_move(MEM, EDX)], EDX}
	end,
    {GrowStack ++ Pushes ++ FixFun1 ++
     store_moves(Stores, FPoff2,
		 simple_moves(Simple, FPoff2,
			      simple_moves(MoreSimple, FPoff2,
					   []))),
     FPoff2, Fun1}.

mk_moves([Arg|Args], Off, Moves) ->
    Off1 = Off - 4,
    mk_moves(Args, Off1, [{Arg,Off1}|Moves]);
mk_moves([], _, Moves) ->
    Moves.

split_moves([Move|Moves], Context, DangerOff, Stores, Simple, Conflict) ->
    {Src,DstOff} = Move,
    case src_is_pseudo(Src) of
	false ->
	    split_moves(Moves, Context, DangerOff, [Move|Stores],
			Simple, Conflict);
	true ->
	    SrcOff = context_offset(Context, Src),
	    Type = typeof_src(Src),
	    if SrcOff =:= DstOff ->
		    split_moves(Moves, Context, DangerOff, Stores,
				Simple, Conflict);
	       SrcOff >= DangerOff ->
		    split_moves(Moves, Context, DangerOff, Stores,
				Simple, [{SrcOff,DstOff,Type}|Conflict]);
	       true ->
		    split_moves(Moves, Context, DangerOff, Stores,
				[{SrcOff,DstOff,Type}|Simple], Conflict)
	    end
    end;
split_moves([], _, _, Stores, Simple, Conflict) ->
    {Stores, Simple, Conflict}.

split_conflict([{SrcOff,DstOff,Type}|Conflict], FPoff, Pushes, Simple) ->
    Push = hipe_x86:mk_push(hipe_x86:mk_mem(mk_esp(), hipe_x86:mk_imm(FPoff+SrcOff), Type)),
    split_conflict(Conflict, FPoff+4, [Push|Pushes], [{-(FPoff+4),DstOff,Type}|Simple]);
split_conflict([], FPoff, Pushes, Simple) ->
    {lists:reverse(Pushes), Simple, FPoff}.

simple_moves([{SrcOff,DstOff,Type}|Moves], FPoff, Rest) ->
    EAX = mk_eax(Type),
    ESP = mk_esp(),
    LoadOff = hipe_x86:mk_imm(FPoff+SrcOff),
    LD = hipe_x86:mk_move(hipe_x86:mk_mem(ESP, LoadOff, Type), EAX),
    StoreOff = hipe_x86:mk_imm(FPoff+DstOff),
    ST = hipe_x86:mk_move(EAX, hipe_x86:mk_mem(ESP, StoreOff, Type)),
    simple_moves(Moves, FPoff, [LD, ST | Rest]);
simple_moves([], _, Rest) ->
    Rest.

store_moves([{Src,DstOff}|Moves], FPoff, Rest) ->
    Type = typeof_src(Src),
    ESP = mk_esp(),
    StoreOff = hipe_x86:mk_imm(FPoff+DstOff),
    ST = hipe_x86:mk_move(Src, hipe_x86:mk_mem(ESP, StoreOff, Type)),
    store_moves(Moves, FPoff, [ST | Rest]);
store_moves([], _, Rest) ->
    Rest.

%%%
%%% Contexts
%%% XXX: use a ref cell in the Context for recording max stack
%%%

-record(context, {liveness, formals, temps, map, ra}).

mk_context(Liveness, Formals, Temps) ->
    RA = hipe_x86:mk_new_temp('untagged'),
    Map = mk_temp_map(Formals, RA, Temps),
    Context = #context{liveness=Liveness, formals=Formals, temps=Temps,
		       map=Map, ra=RA},
    put(x86_maxstack, context_framesize(Context)),
    Context.

context_need_stack(C, N) ->
    M = get(x86_maxstack),
    if N > M ->
	    put(x86_maxstack, N);
       true ->
	    []
    end.

context_maxstack(C) ->
    get(x86_maxstack).

context_arity(#context{formals=Formals}) ->
    length(Formals).

context_framesize(#context{temps=Temps}) ->
    4 * length(Temps).

context_liveness(#context{liveness=Liveness}) ->
    Liveness.

context_offset(#context{map=Map}, Temp) ->
    tmap_lookup(Map, Temp).

context_ra(#context{ra=RA}) ->
    RA.

context_temps(#context{temps=Temps}) ->
    Temps.

mk_temp_map(Formals, RA, Temps) ->
    enter_vars(Temps, -4,
	       enter_vars(Formals, 4*length(Formals),
			  tmap_bind(tmap_empty(), RA, 0))).

enter_vars([V|Vs], Off, Map) ->
    enter_vars(Vs, Off-4, tmap_bind(Map, V, Off));
enter_vars([], _, Map) ->
    Map.

tmap_empty() ->
    orddict:new().

tmap_bind(Map, Key, Val) ->
    orddict:store(Key, Val, Map).

tmap_lookup(Map, Key) ->
    orddict:fetch(Key, Map).

%%%
%%% do_prologue: prepend stack frame allocation code.
%%%
%%% NewStart:
%%%	eax = esp
%%%	eax -= MaxStack
%%%	if( eax < ESP_LIMIT(P) ) goto IncStack else goto AllocFrame
%%% AllocFrame:
%%%	esp -= FrameSize
%%%	goto OldStart
%%% OldStart:
%%%	...
%%% IncStack:
%%%	call inc_stack
%%%	goto NewStart

do_prologue(CFG, Context) ->
    MaxStack = context_maxstack(Context),
    if MaxStack > 0 ->
	    FrameSize = context_framesize(Context),
	    OldStartLab = hipe_x86_cfg:start_label(CFG),
	    NewStartLab = hipe_gensym:get_next_label(x86),
	    AllocFrameLab = hipe_gensym:get_next_label(x86),
	    IncStackLab = hipe_gensym:get_next_label(x86),
	    %%
	    Type = 'untagged',
	    Preg = hipe_x86_registers:proc_pointer(),
	    Pbase = hipe_x86:mk_temp(Preg, Type),
	    ESP_LIMIT_OFF = hipe_x86:mk_imm(hipe_x86_registers:esp_limit_offset()),
	    EAX = mk_eax(Type),
	    ESP = mk_esp(),
	    NewStartCode =
		%% hopefully this lea is faster than the mov;sub it replaced
		[hipe_x86:mk_lea(hipe_x86:mk_mem(ESP, hipe_x86:mk_imm(-MaxStack), 'untagged'), EAX),
		 hipe_x86:mk_cmp(hipe_x86:mk_mem(Pbase, ESP_LIMIT_OFF, Type), EAX),
		 hipe_x86:mk_pseudo_jcc('b', IncStackLab, AllocFrameLab, 0.01)],
	    %%
	    AllocFrameCode =
		[hipe_x86:mk_alu('sub', hipe_x86:mk_imm(FrameSize), ESP),
		 hipe_x86:mk_jmp_label(OldStartLab)],
	    %%
	    IncStackCode =
		[hipe_x86:mk_call(hipe_x86:mk_prim('inc_stack_0')),
		 hipe_x86:mk_jmp_label(NewStartLab)],
	    %%
	    {LLo,_} = hipe_x86_cfg:label_range(CFG),
	    LHi = hipe_gensym:get_label(x86),
	    CFG0 = hipe_x86_cfg:label_range_update(CFG, {LLo,LHi}),
	    %%
	    CFG1 = hipe_x86_cfg:bb_add(CFG0, NewStartLab, hipe_bb:mk_bb(NewStartCode)),
	    CFG2 = hipe_x86_cfg:bb_add(CFG1, AllocFrameLab, hipe_bb:mk_bb(AllocFrameCode)),
	    CFG3 = hipe_x86_cfg:bb_add(CFG2, IncStackLab, hipe_bb:mk_bb(IncStackCode)),
	    CFG4 = hipe_x86_cfg:start_label_update(CFG3, NewStartLab),
	    %%
	    CFG4;
       true ->
	    CFG
    end.

%%% typeof_src -- what's src's type?

typeof_src(Src) ->
    case Src of
	#x86_imm{} ->
	    'untagged';
	#x86_temp{} ->
	    hipe_x86:temp_type(Src);
	#x86_mem{} ->
	    hipe_x86:mem_type(Src)
    end.

%%% Cons up an '%esp' Temp.

mk_esp() ->
    hipe_x86:mk_temp(hipe_x86_registers:esp(), 'untagged').

%%% Cons up an '%eax' Temp.

mk_eax(Type) ->
    hipe_x86:mk_temp(hipe_x86_registers:eax(), Type).

%%% Cons up an '%edx' Temp.

mk_edx() ->
    hipe_x86:mk_temp(hipe_x86_registers:edx(), 'untagged').

%%% Check if an operand is a pseudo-Temp.

src_is_pseudo(Src) ->
    opnd_is_pseudo(Src).

opnd_is_pseudo(Opnd) ->
    case hipe_x86:is_temp(Opnd) of
	true -> temp_is_pseudo(Opnd);
	false -> false
    end.

temp_is_pseudo(Temp) ->
    not(hipe_x86_registers:is_precoloured(hipe_x86:temp_reg(Temp))).

%%%
%%% Build the set of all temps used in a Defun's body.
%%%

all_temps(Code, Formals) ->
    S0 = find_temps(Code, tset_empty()),
    S1 = tset_del_list(S0, Formals),
    S2 = tset_filter(S1, fun(T) -> temp_is_pseudo(T) end),
    tset_to_list(S2).
    
find_temps([I|Insns], S0) ->
    S1 = tset_add_list(S0, hipe_x86_defuse:insn_def(I)),
    S2 = tset_add_list(S1, hipe_x86_defuse:insn_use(I)),
    find_temps(Insns, S2);
find_temps([], S) ->
    S.

tset_empty() ->
    ordsets:new().

tset_add_list(S, Ts) ->
    ordsets:union(S, ordsets:from_list(Ts)).

tset_del_list(S, Ts) ->
    ordsets:subtract(S, ordsets:from_list(Ts)).

tset_filter(S, F) ->
    ordsets:filter(F, S).

tset_to_list(S) ->
    ordsets:to_list(S).

%%% <butt-ugly workaround for "vanishing catch blocks" problem>

init_exnlabs() ->
    put(x86_exnlabs, ordsets:new()).

add_exnlab(L) ->
    put(x86_exnlabs, ordsets:add_element(L, get(x86_exnlabs))).

finalise_exnlabs(CFG) ->
    finalise_exnlabs(ordsets:to_list(get(x86_exnlabs)), CFG).

finalise_exnlabs([], CFG) -> CFG;
finalise_exnlabs([L|Ls], CFG) ->
    finalise_exnlabs(Ls, hipe_x86_cfg:add_fail_entrypoint(CFG, L)).

%%% </butt-ugly workaround for "vanishing catch blocks" problem>
