%%% Coalescing register allocator for x86
%%%
%%% Rules:
%%% - Before RA, every Temp (precoloured or pseudo) is semantically
%%%   equivalent to Reg. Any operand may be Temp.
%%% - Before RA, only FIXED registers may occur in precoloured Temps.
%%%   The exception is move: src or dst may be %eax.
%%% - After RA, an operand (src or dst) may refer to at most one memory cell.
%%%   Therefore, a pseudo-Temp MAY NOT occur as base or offset in an
%%%   explicit memory operand after RA.
%%% - After RA, a binary operation (alu, cmp, move) may refer to at most
%%%   one memory cell. Therefore, AT MOST ONE of src and dst may be a
%%%   pseudo-Temp after RA.
%%% - After RA, the return values in call and return must be the
%%%   %eax and %edx registers.
%%% - After RA, the index in a jmp_switch must be a register.

-module(hipe_x86_ra_coalescing).
-export([ra/1]).
-include("hipe_x86.hrl").
-include("../main/hipe.hrl").

-define(countspills,true).

-ifdef(countspills).
-define(count_spills(X), X).
-define(update_spillsum,
	case get(totalspill) of 
	  {__Loads,__Stores} ->
	    put(totalspill,{__Loads+get(loads),
			    __Stores+get(stores)}),
	    case get(spilledtemps) of
	      __Temps -> 
		put(spilledtemps, __Temps+get(temps));
	      _ ->
		true
	    end;
	  _ -> true 
	end).
-else.
-define(count_spills(X), true).
-define(update_spillsum,true).
-endif.

ra(X86Defun) ->
    ?count_spills({put(loads,0),put(stores,0),put(temps,0)}),
    #defun{code=Code0,label_range=Labels0} = X86Defun,

    CFG1 = hipe_x86_cfg:init(X86Defun),
    hipe_gensym:set_label(x86, hipe_gensym:get_label(rtl)),
    {_, SpillLimit} = hipe_x86_cfg:var_range(CFG1),
    NewCFG = alloc(CFG1, unspilled, 0, SpillLimit, 0),
    Code1 = hipe_x86_cfg_linearise( NewCFG ),
    {LabelsLo,_} = Labels0,
    LabelsHi = hipe_gensym:get_label(x86),
    Labels1 = {LabelsLo,LabelsHi},
    X86Defun#defun{code=Code1,label_range=Labels1}.


alloc( X86Cfg, PrevSpillArea, SpillIndex, SpillLimit, Locs) ->
    {Coloring, NewSpillIndex} = 
	hipe_coalescing_regalloc:regalloc(X86Cfg, SpillIndex, SpillLimit, hipe_x86_specific ),

     ?count_spills(put(temps, NewSpillIndex)),
    case spilled(Coloring) of
	false ->
	    ?update_spillsum,
	    ColTuple = cols2tuple(Coloring),
	    Labels = hipe_x86_cfg:labels(X86Cfg),
	    CFG0 = rewrite_bbs(Labels, X86Cfg, ColTuple),
	    NewCFG = hipe_x86_cfg:var_range_update(CFG0, {0, 31}),
	    case SpillIndex of
		0 ->
		    NewCFG;
		_ ->
		    %% io:format("Spilled: ~w register(s) @ ~w sites (~w)~n", 
		    %%	 [SpillIndex, Locs, hipe_x86_cfg:function(X86Cfg)]),
		    NewConstTab =
			hipe_consttab:update_block(hipe_x86_cfg:data(NewCFG),
						   PrevSpillArea,
						   4,
						   word,
						   hipe_consttab:repeat(SpillIndex, 0)),
		    hipe_x86_cfg:update_data(NewCFG, NewConstTab)
	    end;
	true ->
	    Labels = hipe_x86_cfg:labels(X86Cfg),
	    {Low, High} = hipe_x86_cfg:var_range(X86Cfg),
	    hipe_gensym:set_var(High),
	    {CFG1, SpillArea} = 
		case PrevSpillArea of
		    unspilled ->
			{NewCTab, SpillA} = 
			    hipe_consttab:insert_block(hipe_x86_cfg:data(X86Cfg),
						       4, word, [0]),
			{hipe_x86_cfg:update_data(X86Cfg, NewCTab), SpillA};
		    SpillA ->
			{X86Cfg, SpillA}
		end,
	    
	    Spills = spill_regs(Coloring),
	    {CFG2, Locs0} = spill_rewrite_bbs(Labels, CFG1, Spills, 
					      SpillArea, Locs),
	    CFG3 = hipe_x86_cfg:var_range_update(CFG2,
						   {Low, hipe_gensym:get_var()}),
	    alloc(CFG3, SpillArea, NewSpillIndex, SpillLimit,Locs0)
    end.

%
% Check if a coloring spilled any values
%

spilled([]) ->
   false;
spilled([{_, {spill, _}} | Cs]) ->
   true;
spilled([_ | Cs]) ->
   spilled(Cs).

%
% Returns a list of {SpilledReg, {Tmp, LoadInstr, StoreInstr}}
%

spill_regs([]) ->
   [];
spill_regs([{RegNr, {reg, _}} | Colors]) ->
   spill_regs(Colors);
spill_regs([{RegNr, {spill, SpillIndex}} | Colors]) ->
   SpillIndexImm = hipe_x86:mk_imm(SpillIndex*4),
   Info = {hipe_x86:mk_reg(RegNr), SpillIndexImm},
   [Info | spill_regs(Colors)].


%
% Rewrite a cfg where spills occured
%

spill_rewrite_bbs([], CFG, Spills, SpillArea, Locs) ->
   {CFG, Locs};
spill_rewrite_bbs([Lbl|Lbls], CFG, Spills, SpillArea, Locs) ->
   BB = hipe_x86_cfg:bb(CFG, Lbl),
   Code = hipe_bb:code(BB),
   {NewCode, Locs0} = spill_rewrite_instrs(Code, Spills, SpillArea),
   NewCFG = hipe_x86_cfg:bb_update(CFG, Lbl, hipe_bb:code_update(BB, NewCode)),
   spill_rewrite_bbs(Lbls, NewCFG, Spills, SpillArea, Locs+Locs0).

%
% Convert a list of [{R1, C1}, {R2, C2}, ...} to a tuple {C17, C23, ...}.
%
% The N's must be unique but do not have to be sorted and they can be sparse.
%

cols2tuple(Map) ->
   cols2tuple(1, lists:keysort(1, Map), []).

cols2tuple(N, [], Vs) ->
   list_to_tuple(lists:reverse(Vs));
cols2tuple(N, [{R, C}|Ms], Vs) when N =:= R ->
   cols2tuple(N+1, Ms, [C|Vs]);
cols2tuple(N, Ms, Vs) ->
   cols2tuple(N+1, Ms, [unknown|Vs]).


%
% Rewrite a cfg to use the allocated registers
%

rewrite_bbs([], CFG, ColTuple) ->
   CFG;
rewrite_bbs([Lbl|Lbls], CFG, ColTuple) ->
   BB = hipe_x86_cfg:bb(CFG, Lbl),
   Code = hipe_bb:code(BB),
   NewCode = rewrite_instrs(Code, ColTuple),
   NewCFG = hipe_x86_cfg:bb_update(CFG, Lbl, hipe_bb:code_update(BB, NewCode)),
   rewrite_bbs(Lbls, NewCFG, ColTuple).

spill_rewrite_instrs([], Spills, SpillArea) ->
   {[], 0};
spill_rewrite_instrs([I|Is], Spills, SpillArea) ->
   {NewI, Locs0} = spill_rewrite_instr(I, Spills, SpillArea),
   {NewIs, Locs1} = spill_rewrite_instrs(Is, Spills, SpillArea),
   {NewI++NewIs, Locs0+Locs1}.


spill_rewrite_instr(I, Spills, SpillArea) ->
    {Defs, Uses} = hipe_x86:def_use(I),
    SpillUses = get_spills(Uses, Spills, Spills),
    SpillDefs = get_spills(Defs, Spills, Spills),
    case SpillUses++SpillDefs of
	[] ->
	    {[I], 0};
	_ ->
	    SpillAreaReg = hipe_x86:mk_new_reg(),
	    LA = hipe_x86:load_address_create(SpillAreaReg,
					   SpillArea, constant, []),
	    {Loads,UseSubst} = 
		case SpillUses of
		    [] ->
			{[],[]};
		    [{R1u, Offset1u}] ->
			Tmp1u = hipe_x86:mk_new_reg(),
			{[hipe_x86:load_create(Tmp1u, uw,
					    SpillAreaReg, Offset1u, [])],
			 [{R1u, Tmp1u}]};
		    [{R1u, Offset1u}, {R2u, Offset2u}] ->
			Tmp1u = hipe_x86:mk_new_reg(),
			Tmp2u = hipe_x86:mk_new_reg(),
			{[hipe_x86:load_create(Tmp1u, uw,
					    SpillAreaReg, Offset1u, []),
			  hipe_x86:load_create(Tmp2u, uw,
					    SpillAreaReg, Offset2u, [])],
			 [{R1u, Tmp1u}, {R2u, Tmp2u}]}
		end,
	    {Stores,DefSubst} = 
		case SpillDefs of
		    [] ->
			{[],[]};
		    [{R1d, Offset1d}] ->
			Tmp1d = hipe_x86:mk_new_reg(),
			{[hipe_x86:store_create(SpillAreaReg,
					     Offset1d, uw, Tmp1d, [])],
			 [{R1d, Tmp1d}]}
		end,
       ?count_spills(
	  {put(stores,get(stores)+length(Stores)),
	   put(loads,get(loads)+length(Loads))}),
	 C1 = hipe_x86:comment_create('** SPILL START **', []),
	 C2 = hipe_x86:comment_create('** SPILL END **', []),
	 NewI = hipe_x86:subst_defines(hipe_x86:subst_uses(I, UseSubst), DefSubst),
	 {[C1, LA] ++ Loads ++ [NewI] ++ Stores ++ [C2], length(UseSubst)+length(DefSubst)}
   end.

rewrite_instrs([], ColTuple) ->
   [];
rewrite_instrs([I|Is], ColTuple) ->
   [rewrite_instr(I, ColTuple) | rewrite_instrs(Is, ColTuple)].


rewrite_instr(Ins, ColTuple) ->
   case hipe_x86:type(Ins) of
      label -> Ins;
      nop -> Ins;
      block -> Ins;
      align -> Ins;
      comment -> Ins;
      b -> Ins;
      goto -> Ins;
      move ->
	 NewDst = color_arg(hipe_x86:move_dest(Ins), ColTuple),
	 NewSrc = color_arg(hipe_x86:move_src(Ins), ColTuple),
	 hipe_x86:move_dest_update(hipe_x86:move_src_update(Ins, NewSrc), NewDst);
      cmov_cc ->
	 NewDst = color_arg(hipe_x86:cmov_cc_dest(Ins), ColTuple),
	 NewSrc = color_arg(hipe_x86:cmov_cc_src(Ins), ColTuple),
	 I0 = hipe_x86:cmov_cc_src_update(Ins, NewSrc),
	 hipe_x86:cmov_cc_dest_update(I0, NewDst);
      cmov_r ->
	 NewDst = color_arg(hipe_x86:cmov_r_dest(Ins), ColTuple),
	 NewSrc = color_arg(hipe_x86:cmov_r_src(Ins), ColTuple),
	 NewReg = color_arg(hipe_x86:cmov_r_reg(Ins), ColTuple),
	 I0 = hipe_x86:cmov_r_dest_update(Ins, NewDst),
	 I1 = hipe_x86:cmov_r_src_update(I0, NewSrc),
	 hipe_x86:cmov_r_reg_update(I1, NewReg);
      alu ->
	 NewSrc1 = color_arg(hipe_x86:alu_src1(Ins), ColTuple),
	 NewSrc2 = color_arg(hipe_x86:alu_src2(Ins), ColTuple),
	 NewDst = color_arg(hipe_x86:alu_dest(Ins), ColTuple),
	 I0 = hipe_x86:alu_src1_update(Ins, NewSrc1),
	 I1 = hipe_x86:alu_src2_update(I0, NewSrc2),
	 hipe_x86:alu_dest_update(I1, NewDst);
      alu_cc ->
	 NewSrc1 = color_arg(hipe_x86:alu_cc_src1(Ins), ColTuple),
	 NewSrc2 = color_arg(hipe_x86:alu_cc_src2(Ins), ColTuple),
	 NewDst = color_arg(hipe_x86:alu_cc_dest(Ins), ColTuple),
	 I0 = hipe_x86:alu_cc_src1_update(Ins, NewSrc1),
	 I1 = hipe_x86:alu_cc_src2_update(I0, NewSrc2),
	 hipe_x86:alu_cc_dest_update(I1, NewDst);
      sethi ->
	 NewDest = color_arg(hipe_x86:sethi_dest(Ins), ColTuple),
	 hipe_x86:sethi_dest_update(Ins, NewDest);

      load ->
	 NewDest = color_arg(hipe_x86:load_dest(Ins), ColTuple),
	 NewSrc = color_arg(hipe_x86:load_src(Ins), ColTuple),
	 NewOff = color_arg(hipe_x86:load_off(Ins), ColTuple),
	 I0 = hipe_x86:load_dest_update(Ins, NewDest),
	 I1 = hipe_x86:load_src_update(I0, NewSrc),
	 hipe_x86:load_off_update(I1, NewOff);
      store ->
	 NewDest = color_arg(hipe_x86:store_dest(Ins), ColTuple),
	 NewSrc = color_arg(hipe_x86:store_src(Ins), ColTuple),
	 NewOff = color_arg(hipe_x86:store_off(Ins), ColTuple),
	 I0 = hipe_x86:store_dest_update(Ins, NewDest),
	 I1 = hipe_x86:store_src_update(I0, NewSrc),
	 hipe_x86:store_off_update(I1, NewOff);
      br ->
	 NewReg = color_arg(hipe_x86:br_reg(Ins), ColTuple),
	 hipe_x86:br_reg_update(Ins, NewReg);
      %% Warning, not complete
      jmp_link ->
	 NewTarget = color_arg(hipe_x86:jmp_link_target(Ins), ColTuple),
	 NewOff = color_arg(hipe_x86:jmp_link_off(Ins), ColTuple),
	 I0 = hipe_x86:jmp_link_target_update(Ins, NewTarget),
	 hipe_x86:jmp_link_off_update(I0, NewOff);
      jmp ->
	 NewTarget = color_arg(hipe_x86:jmp_target(Ins), ColTuple),
	 NewOff = color_arg(hipe_x86:jmp_off(Ins), ColTuple),
	 I0 = hipe_x86:jmp_target_update(Ins, NewTarget),
	 hipe_x86:jmp_off_update(I0, NewOff);
      call_link ->
         Ins1 =
         case hipe_x86:call_link_type(Ins) of
	   closure ->
	     NewTarget = color_arg(hipe_x86:call_link_target(Ins), ColTuple),
	     hipe_x86:call_link_target_update(Ins, NewTarget);
	   _ -> Ins
	 end,
	 NewLink = color_arg(hipe_x86:call_link_link(Ins1), ColTuple),
	 hipe_x86:call_link_link_update(Ins1, NewLink);
      %% end warning
      load_atom ->
	 NewDest = color_arg(hipe_x86:load_atom_dest(Ins), ColTuple),
	 hipe_x86:load_atom_dest_update(Ins, NewDest);
      load_address ->
	 NewDest = color_arg(hipe_x86:load_address_dest(Ins), ColTuple),
	 hipe_x86:load_address_dest_update(Ins, NewDest);
      _ -> Ins
   end.


color_arg(Arg, ColTuple) ->
   case hipe_x86:is_reg(Arg) of
      true ->
	 case element(hipe_x86:reg_nr(Arg), ColTuple) of
	    {reg, NewRgNr} ->
	       hipe_x86:mk_reg(NewRgNr);
	    {spill, SpillIndex} ->
	       exit({x86, spilled})
	 end;
      false ->
	 Arg
   end.













do_insns([I|Insns]) ->
    do_insn(I) ++ do_insns(Insns);
do_insns([]) ->
    [].

do_insn(I) ->	% Insn -> Insn list
    case I of
	#alu{} ->
	    do_alu(I);
	#cmp{} ->
	    do_cmp(I);
	#jmp_switch{} ->
	    do_jmp_switch(I);
	#move{} ->
	    do_move(I);
	#pseudo_call{} ->
	    do_pseudo_call(I);
	#pseudo_return{} ->
	    do_pseudo_return(I);
	_ ->    %% comment, jmp*, label, pseudo_jcc, pseudo_tailcall
	    [I]
    end.

%%% Fix an alu op.

do_alu(I) ->
    #alu{src=Src0,dst=Dst0} = I,
    {FixSrc,Src,FixDst,Dst} = do_binary(Src0, Dst0),
    FixSrc ++ FixDst ++ [I#alu{src=Src,dst=Dst}].

%%% Fix a cmp op.

do_cmp(I) ->
    #cmp{src=Src0,dst=Dst0} = I,
    {FixSrc, Src, FixDst, Dst} = do_binary(Src0, Dst0),
    FixSrc ++ FixDst ++ [I#cmp{src=Src,dst=Dst}].

%%% Fix a jmp_switch op.

do_jmp_switch(I) ->
    #jmp_switch{temp=Temp} = I,
    case temp_is_pseudo(Temp) of
	false ->
	    [I];
	true ->
	    EAX = hipe_x86:mk_temp(hipe_x86_registers:eax(), 'untagged'),
	    [hipe_x86:mk_move(Temp, EAX), I#jmp_switch{temp=EAX}]
    end.

%%% Fix a move op.

do_move(I) ->
    #move{src=Src0,dst=Dst0} = I,
    {FixSrc, Src, FixDst, Dst} = do_binary(Src0, Dst0),
    FixSrc ++ FixDst ++ [I#move{src=Src,dst=Dst}].

%%% Move call-defined registers to the result temps.
%%% Calls normally return one result, but many BIFs return two (alas),
%%% and some return none (suspend/0).

do_pseudo_call(I) ->
    #pseudo_call{dsts=Dsts, contlab=ContLab} = I,
    case do_call_results(Dsts) of
	{[], []} ->
	    %% Avoid consing up a dummy basic block if the moves list
	    %% is empty, as is typical for calls to suspend/0.
	    %% This would be subsumed by a general "optimise the CFG"
	    %% module, but we don't have that one yet :-(
	    [I];
	{Moves, Dsts2} ->
	    NewLab = hipe_gensym:get_next_label(x86),
	    [I#pseudo_call{dsts=Dsts2, contlab=NewLab},
	     hipe_x86:mk_label(NewLab, false) |
	     Moves ++
	     [hipe_x86:mk_jmp_label(ContLab)]]
    end.

do_call_results([Dst]) ->
    EAX = hipe_x86:mk_temp(hipe_x86_registers:eax(), 'tagged'),
    Move = hipe_x86:mk_move(EAX, Dst),	% Dst <- %eax
    {[Move], [EAX]};
do_call_results([Dst, Flag]) ->	% XXX: shouldn't happen any more
    EAX = hipe_x86:mk_temp(hipe_x86_registers:eax(), 'tagged'),
    Move1 = hipe_x86:mk_move(EAX, Dst),
    EDX = hipe_x86:mk_temp(hipe_x86_registers:edx(), 'untagged'),
    Move2 = hipe_x86:mk_move(EDX, Flag),
    {[Move1,Move2], [EAX,EDX]};
do_call_results([]) ->
    {[], []}.

%%% Put the return values in their proper physical registers.
%%% A compiled Erlang function returns a single value.
%%% (Many BIFs currently return two values, unfortunately.)

do_pseudo_return(I) ->
    #pseudo_return{args=Args} = I,
    case Args of
	[Arg] ->
	    EAX = hipe_x86:mk_temp(hipe_x86_registers:eax(), 'tagged'),
	    MV = hipe_x86:mk_move(Arg, EAX),
	    [MV, I#pseudo_return{args=[EAX]}]
	%% deal with 0 or 2 retvals here, if necessary
    end.

%%% Fix the operands of a binary op.
%%% 1. remove pseudos from any explicit memory operands
%%% 2. if both operands are (implicit or explicit) memory operands,
%%%    move src to a reg and use reg as src in the original insn

do_binary(Src0, Dst0) ->
    {FixSrc, Src} = fix_src_operand(Src0),
    {FixDst, Dst} = fix_dst_operand(Dst0),
    {FixSrc3, Src3} =
	case is_mem_opnd(Src) of
	    false ->
		{FixSrc, Src};
	    true ->
		case is_mem_opnd(Dst) of
		    false ->
			{FixSrc, Src};
		    true ->
			EAX = hipe_x86_registers:eax(),
			Src2 = clone(Src, EAX),
			FixSrc2 = FixSrc ++ [hipe_x86:mk_move(Src, Src2)],
			{FixSrc2, Src2}
		end
	end,
    {FixSrc3, Src3, FixDst, Dst}.

%%% Fix any x86_mem operand to not refer to any pseudos.
%%% The fixup may use additional instructions and registers.
%%% 'src' operands may clobber '%eax'.
%%% 'dst' operands may clobber '%edx'.
%%% (This is not arbitrary. The translation of restore_catch does a
%%% "move %eax,<dst>", and fail_to does a "move <src>,%eax".)

fix_src_operand(Opnd) ->
    fix_mem_operand(Opnd, hipe_x86_registers:eax()).

fix_dst_operand(Opnd) ->
    fix_mem_operand(Opnd, hipe_x86_registers:edx()).

fix_mem_operand(Opnd, Reg) ->	% -> {[fixupcode], newop}
    case Opnd of
	#x86_mem{base=Base,off=Off} ->
	    case is_mem_opnd(Base) of
		false ->
		    case src_is_pseudo(Off) of
			false ->
			    {[], Opnd};
			true ->		% pseudo(reg)
			    Temp = clone(Off, Reg),
			    {[hipe_x86:mk_move(Off, Temp)],
			     Opnd#x86_mem{off=Temp}}
		    end;
		true ->
		    Temp = clone(Base, Reg),
		    case src_is_pseudo(Off) of
			false ->	% imm/reg(pseudo)
			    {[hipe_x86:mk_move(Base, Temp)],
			     Opnd#x86_mem{base=Temp}};
			true ->		% pseudo1(pseudo0)
			    {[hipe_x86:mk_move(Base, Temp),
			      hipe_x86:mk_alu('add', Off, Temp)],
			     Opnd#x86_mem{base=Temp, off=hipe_x86:mk_imm(0)}}
		    end
	    end;
	_ ->
	    {[], Opnd}
    end.

%%% Check if an operand denotes a memory cell (mem or pseudo).

is_mem_opnd(Opnd) ->
    case Opnd of
	#x86_mem{} -> true;
	#x86_temp{} -> temp_is_pseudo(Opnd);
	_ -> false
    end.

%%% Check if an operand is a pseudo-Temp.

src_is_pseudo(Src) ->
    case hipe_x86:is_temp(Src) of
	true -> temp_is_pseudo(Src);
	false -> false
    end.

temp_is_pseudo(Temp) ->
    not(hipe_x86_registers:is_precoloured(hipe_x86:temp_reg(Temp))).

%%% Make Reg a clone of Dst (attach Dst's type to Reg).

clone(Dst, Reg) ->
    Type =
	case Dst of
	    #x86_mem{} -> hipe_x86:mem_type(Dst);
	    #x86_temp{} -> hipe_x86:temp_type(Dst)
	end,
    hipe_x86:mk_temp(Reg, Type).
