%%% $Id: hipe_x86_ra_dummy.erl,v 1.12 2001/06/13 12:55:10 mikpe Exp $
%%% simple local x86 regalloc
%%%
%%% Rules:
%%% - Before RA, every Temp (precoloured or pseudo) is semantically
%%%   equivalent to Reg. Any operand may be Temp.
%%% - Before RA, only FIXED registers may occur in precoloured Temps.
%%%   The exception is move: src or dst may be %eax.
%%% - After RA, an operand (src or dst) may refer to at most one memory cell.
%%%   Therefore, a pseudo-Temp MAY NOT occur as base or offset in an
%%%   explicit memory operand after RA.
%%% - After RA, a binary operation (alu, cmp, move) may refer to at most
%%%   one memory cell. Therefore, AT MOST ONE of src and dst may be a
%%%   pseudo-Temp after RA.
%%% - After RA, the return values in call and return must be the
%%%   %eax and %edx registers.
%%% - After RA, the index in a jmp_switch must be a register.
%%% - After RA, the temp in a lea must be a register.

-module(hipe_x86_ra_dummy).
-export([ra/1]).
-include("hipe_x86.hrl").

ra(X86Defun) ->
    #defun{code=Code0,label_range=Labels0} = X86Defun,
    hipe_gensym:set_label(x86, hipe_gensym:get_label(rtl)),
    Code1 = do_insns(Code0),
    {LabelsLo,_} = Labels0,
    LabelsHi = hipe_gensym:get_label(x86),
    Labels1 = {LabelsLo,LabelsHi},
    X86Defun#defun{code=Code1,label_range=Labels1}.

do_insns([I|Insns]) ->
    do_insn(I) ++ do_insns(Insns);
do_insns([]) ->
    [].

do_insn(I) ->	% Insn -> Insn list
    case I of
	#alu{} ->
	    do_alu(I);
	#cmp{} ->
	    do_cmp(I);
	#jmp_switch{} ->
	    do_jmp_switch(I);
	#lea{} ->
	    do_lea(I);
	#move{} ->
	    do_move(I);
	#pseudo_call{} ->
	    do_pseudo_call(I);
	#pseudo_return{} ->
	    do_pseudo_return(I);
	_ ->    %% comment, jmp*, label, pseudo_jcc, pseudo_tailcall
	    [I]
    end.

%%% Fix an alu op.

do_alu(I) ->
    #alu{src=Src0,dst=Dst0} = I,
    {FixSrc,Src,FixDst,Dst} = do_binary(Src0, Dst0),
    FixSrc ++ FixDst ++ [I#alu{src=Src,dst=Dst}].

%%% Fix a cmp op.

do_cmp(I) ->
    #cmp{src=Src0,dst=Dst0} = I,
    {FixSrc, Src, FixDst, Dst} = do_binary(Src0, Dst0),
    FixSrc ++ FixDst ++ [I#cmp{src=Src,dst=Dst}].

%%% Fix a jmp_switch op.

do_jmp_switch(I) ->
    #jmp_switch{temp=Temp} = I,
    case temp_is_pseudo(Temp) of
	false ->
	    [I];
	true ->
	    EAX = hipe_x86:mk_temp(hipe_x86_registers:eax(), 'untagged'),
	    [hipe_x86:mk_move(Temp, EAX), I#jmp_switch{temp=EAX}]
    end.

%%% Fix a lea op.

do_lea(I) ->
    #lea{temp=Temp} = I,
    case temp_is_pseudo(Temp) of
	false ->
	    [I];
	true ->
	    EAX = hipe_x86:mk_temp(hipe_x86_registers:eax(), 'untagged'),
	    [I#lea{temp=EAX}, hipe_x86:mk_move(EAX, Temp)]
    end.

%%% Fix a move op.

do_move(I) ->
    #move{src=Src0,dst=Dst0} = I,
    {FixSrc, Src, FixDst, Dst} = do_binary(Src0, Dst0),
    FixSrc ++ FixDst ++ [I#move{src=Src,dst=Dst}].

%%% Move call-defined registers to the result temps.
%%% Calls normally return one result, but many BIFs return two (alas),
%%% and some return none (suspend/0).

do_pseudo_call(I) ->
    #pseudo_call{dsts=Dsts, contlab=ContLab} = I,
    case do_call_results(Dsts) of
	{[], []} ->
	    %% Avoid consing up a dummy basic block if the moves list
	    %% is empty, as is typical for calls to suspend/0.
	    %% This would be subsumed by a general "optimise the CFG"
	    %% module, but we don't have that one yet :-(
	    [I];
	{Moves, Dsts2} ->
	    NewLab = hipe_gensym:get_next_label(x86),
	    [I#pseudo_call{dsts=Dsts2, contlab=NewLab},
	     hipe_x86:mk_label(NewLab, false) |
	     Moves ++
	     [hipe_x86:mk_jmp_label(ContLab)]]
    end.

do_call_results([Dst]) ->
    EAX = hipe_x86:mk_temp(hipe_x86_registers:eax(), 'tagged'),
    Move = hipe_x86:mk_move(EAX, Dst),	% Dst <- %eax
    {[Move], [EAX]};
do_call_results([Dst, Flag]) ->	% XXX: shouldn't happen any more
    EAX = hipe_x86:mk_temp(hipe_x86_registers:eax(), 'tagged'),
    Move1 = hipe_x86:mk_move(EAX, Dst),
    EDX = hipe_x86:mk_temp(hipe_x86_registers:edx(), 'untagged'),
    Move2 = hipe_x86:mk_move(EDX, Flag),
    {[Move1,Move2], [EAX,EDX]};
do_call_results([]) ->
    {[], []}.

%%% Put the return values in their proper physical registers.
%%% A compiled Erlang function returns a single value.
%%% (Many BIFs currently return two values, unfortunately.)

do_pseudo_return(I) ->
    #pseudo_return{args=Args} = I,
    case Args of
	[Arg] ->
	    EAX = hipe_x86:mk_temp(hipe_x86_registers:eax(), 'tagged'),
	    MV = hipe_x86:mk_move(Arg, EAX),
	    [MV, I#pseudo_return{args=[EAX]}]
	%% deal with 0 or 2 retvals here, if necessary
    end.

%%% Fix the operands of a binary op.
%%% 1. remove pseudos from any explicit memory operands
%%% 2. if both operands are (implicit or explicit) memory operands,
%%%    move src to a reg and use reg as src in the original insn

do_binary(Src0, Dst0) ->
    {FixSrc, Src} = fix_src_operand(Src0),
    {FixDst, Dst} = fix_dst_operand(Dst0),
    {FixSrc3, Src3} =
	case is_mem_opnd(Src) of
	    false ->
		{FixSrc, Src};
	    true ->
		case is_mem_opnd(Dst) of
		    false ->
			{FixSrc, Src};
		    true ->
			EAX = hipe_x86_registers:eax(),
			Src2 = clone(Src, EAX),
			FixSrc2 = FixSrc ++ [hipe_x86:mk_move(Src, Src2)],
			{FixSrc2, Src2}
		end
	end,
    {FixSrc3, Src3, FixDst, Dst}.

%%% Fix any x86_mem operand to not refer to any pseudos.
%%% The fixup may use additional instructions and registers.
%%% 'src' operands may clobber '%eax'.
%%% 'dst' operands may clobber '%edx'.
%%% (This is not arbitrary. The translation of restore_catch does a
%%% "move %eax,<dst>", and fail_to does a "move <src>,%eax".)

fix_src_operand(Opnd) ->
    fix_mem_operand(Opnd, hipe_x86_registers:eax()).

fix_dst_operand(Opnd) ->
    fix_mem_operand(Opnd, hipe_x86_registers:edx()).

fix_mem_operand(Opnd, Reg) ->	% -> {[fixupcode], newop}
    case Opnd of
	#x86_mem{base=Base,off=Off} ->
	    case is_mem_opnd(Base) of
		false ->
		    case src_is_pseudo(Off) of
			false ->
			    {[], Opnd};
			true ->		% pseudo(reg)
			    Temp = clone(Off, Reg),
			    {[hipe_x86:mk_move(Off, Temp)],
			     Opnd#x86_mem{off=Temp}}
		    end;
		true ->
		    Temp = clone(Base, Reg),
		    case src_is_pseudo(Off) of
			false ->	% imm/reg(pseudo)
			    {[hipe_x86:mk_move(Base, Temp)],
			     Opnd#x86_mem{base=Temp}};
			true ->		% pseudo1(pseudo0)
			    {[hipe_x86:mk_move(Base, Temp),
			      hipe_x86:mk_alu('add', Off, Temp)],
			     Opnd#x86_mem{base=Temp, off=hipe_x86:mk_imm(0)}}
		    end
	    end;
	_ ->
	    {[], Opnd}
    end.

%%% Check if an operand denotes a memory cell (mem or pseudo).

is_mem_opnd(Opnd) ->
    case Opnd of
	#x86_mem{} -> true;
	#x86_temp{} -> temp_is_pseudo(Opnd);
	_ -> false
    end.

%%% Check if an operand is a pseudo-Temp.

src_is_pseudo(Src) ->
    case hipe_x86:is_temp(Src) of
	true -> temp_is_pseudo(Src);
	false -> false
    end.

temp_is_pseudo(Temp) ->
    not(hipe_x86_registers:is_precoloured(hipe_x86:temp_reg(Temp))).

%%% Make Reg a clone of Dst (attach Dst's type to Reg).

clone(Dst, Reg) ->
    Type =
	case Dst of
	    #x86_mem{} -> hipe_x86:mem_type(Dst);
	    #x86_temp{} -> hipe_x86:temp_type(Dst)
	end,
    hipe_x86:mk_temp(Reg, Type).
