%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.1, (the "License"); you may not use this file except in
%% compliance with the License. You should have received a copy of the
%% Erlang Public License along with this software. If not, it can be
%% retrieved via the world wide web at http://www.erlang.org/.
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Initial Developer of the Original Code is Ericsson Utvecklings AB.
%% Portions created by Ericsson are Copyright 1999, Ericsson Utvecklings
%% AB. All Rights Reserved.''
%% 
%%     $Id$
%%
%%----------------------------------------------------------------------
%% Purpose: 
%%----------------------------------------------------------------------

-module(megaco_sdp).

%%----------------------------------------------------------------------
%% Include files
%%----------------------------------------------------------------------

-include_lib("megaco/include/megaco.hrl").
-include_lib("megaco/include/megaco_message_v1.hrl").
-include_lib("megaco/include/megaco_sdp.hrl").

%%----------------------------------------------------------------------
%% External exports
%%----------------------------------------------------------------------

-export([
	 parseLRDesc/1,
	 genLRDesc/2,
	 chooseSDPProps/2,
	 resetSDPPropsToChoose/1,
	 connection_addr/1,
	 port/1,
	 payload_type/1
        ]).

%%----------------------------------------------------------------------
%% Internal exports
%%----------------------------------------------------------------------

-export([
        ]).

%%----------------------------------------------------------------------
%% Macros
%%----------------------------------------------------------------------

%%----------------------------------------------------------------------
%% Records
%%----------------------------------------------------------------------

%%======================================================================
%% External functions
%%======================================================================
%%-----------------------------------------------------------------
%% Func: parseLRDesc/1
%% Description: Parse the SDP information from the LocalRemoteDescriptor
%%-----------------------------------------------------------------

parseLRDesc([]) ->
    [];
parseLRDesc(asn1_NOVALUE) ->
    [];
parseLRDesc([PropGrp |PropGrps]) ->
    case parseLRDesc_1(PropGrp, []) of
	no ->
	    parseLRDesc(PropGrps);
	NewPropGrp ->
	    [NewPropGrp | parseLRDesc(PropGrps)]
    end;
parseLRDesc(Descr) 
  when record(Descr,'LocalRemoteDescriptor') ->
    DecodedPropGroups = parseLRDesc(Descr#'LocalRemoteDescriptor'.propGrps),
    Descr#'LocalRemoteDescriptor'{propGrps = DecodedPropGroups}.


parseLRDesc_1([], []) ->
    no;
parseLRDesc_1([], Acc) ->
    lists:reverse(Acc);
parseLRDesc_1([#'PropertyParm'{name="v", value=[V], extraInfo=EI}| PPs], Acc) ->
		     parseLRDesc_1(PPs,[#sdp_v{version=V} | Acc]);
parseLRDesc_1([#'PropertyParm'{name="c", value=[Val], extraInfo=EI}| PPs], Acc) ->
    case string:tokens(Val, " \t") of
	[Network, AddressType, ConnectionAddr]	->
	    parseLRDesc_1(PPs, 
			  [#sdp_c{network_type = Network,
				  address_type = AddressType,
				  connection_addr = ConnectionAddr
				 } 
			   | Acc]);
	_ ->
	    no
    end;
parseLRDesc_1([#'PropertyParm'{name="m", value=[Val], extraInfo=EI}| PPs], Acc) ->
    case string:tokens(Val, " \t") of
	[Media, PortInfo, Transport | FMT]	->
	    {Port, NoOfPorts} = 
		case string:tokens(PortInfo, "/") of
		    [P1, NP] ->
			{list_to_integer(P1), list_to_integer(NP)};
		    [P2] ->
			{list_to_integer(P2), undefined}
		end,
	    parseLRDesc_1(PPs, [#sdp_m{media = Media,
				       port = Port,
				       num_ports = NoOfPorts,
				       transport = Transport,
				       fmt_list = FMT}
				| Acc]);
	_ ->
	    no
    end;
parseLRDesc_1([#'PropertyParm'{name="o", value=[Val], extraInfo=EI}| PPs], Acc) ->
    case string:tokens(Val, " \t") of
	[User, SessionId, Version, Network, AddressType, Address] ->
	    parseLRDesc_1(PPs, [#sdp_o{user_name = User,
				       session_id = SessionId,
				       version = Version,
				       network_type = Network,
				       address_type = AddressType,
				       address = Address}
				| Acc]);
	_ ->
	    no
    end;
parseLRDesc_1([#'PropertyParm'{name="a", value=[Val], extraInfo=EI}| PPs], 
	      Acc) ->
    case string:tokens(Val, ":/ \t") of
	["rtpmap", Payload, EncName, ClockRate | EncPar] ->
	    parseLRDesc_1(PPs, 
			  [#sdp_a_rtpmap{payload_type = 
					 list_to_integer(Payload), 
					 encoding_name = EncName,
					 clock_rate = 
					 list_to_integer(ClockRate),
					 encoding_parms = EncPar}
			   |Acc]);
	["ptime", PacketTime] ->
	    parseLRDesc_1(PPs, 
			  [#sdp_a_ptime{packet_time = 
					list_to_integer(PacketTime)}
			   |Acc]);
	[Var, Value] ->
	    parseLRDesc_1(PPs,[#sdp_a{attribute = Var, value = Value}
			       |Acc]);
	[Var] ->
	    parseLRDesc_1(PPs,[#sdp_a{attribute = Var}
			       |Acc]);
	_ ->
	    no
    end;
parseLRDesc_1([#'PropertyParm'{name="b", value=[Val], extraInfo=EI}| PPs], 
	      Acc) ->
    case string:tokens(Val, ":") of
	[Modifier, Bandwidth] ->
	    parseLRDesc_1(PPs, [#sdp_b{modifier=Modifier,
				       bandwidth=list_to_integer(Bandwidth)} 
				| Acc]);
	_ ->
	    no
    end;
parseLRDesc_1([#'PropertyParm'{name="t", value=[Val], extraInfo=EI}| PPs], 
	      Acc) ->
    case string:tokens(Val, " \t") of
	[Start, Stop] ->    
	    parseLRDesc_1(PPs, [#sdp_t{start=list_to_integer(Start), 
				       stop=list_to_integer(Stop)} | Acc]);
	_ ->
	    no
    end;	
parseLRDesc_1([#'PropertyParm'{name="r", value=[Val], extraInfo=EI}| PPs], 
	      Acc) ->
    case string:tokens(Val, " \t") of
	[Repeat, Duration | ListOfOffsets] ->    
	    parseLRDesc_1(PPs, [#sdp_r{repeat_interval=Repeat,
			       active_duration=Duration, 
			       list_of_offsets=ListOfOffsets} | Acc]);
	_ ->
	    no
    end;	
parseLRDesc_1([#'PropertyParm'{name="z", value=[Val], extraInfo=EI}| PPs], 
	      Acc) ->
    case string:tokens(Val, " \t") of
	List ->    
	    parseLRDesc_1(PPs, [#sdp_z{list_of_adjustments=List} | Acc]);
	_ ->
	    no
    end;	
parseLRDesc_1([#'PropertyParm'{name="k", value=[Val], extraInfo=EI}| PPs], 
	      Acc) ->
    case string:tokens(Val, ":") of
	[Method, EncryptionKey] ->    
	    parseLRDesc_1(PPs, [#sdp_k{method=Method, 
				       encryption_key=EncryptionKey} | Acc]);
	_ ->
	    no
    end;	

parseLRDesc_1([#'PropertyParm'{name="s", value=[Val], extraInfo=EI}| PPs], 
	      Acc) ->
    parseLRDesc_1(PPs, [#sdp_s{name=Val} | Acc]);
parseLRDesc_1([#'PropertyParm'{name="i", value=[Val], extraInfo=EI}| PPs], 
	      Acc) ->
    parseLRDesc_1(PPs, [#sdp_i{session_descriptor=Val} | Acc]);
parseLRDesc_1([#'PropertyParm'{name="u", value=[Val], extraInfo=EI}| PPs], 
	      Acc) ->
    parseLRDesc_1(PPs, [#sdp_u{uri=Val} | Acc]);
parseLRDesc_1([#'PropertyParm'{name="e", value=[Val], extraInfo=EI}| PPs], 
	      Acc) ->
    parseLRDesc_1(PPs, [#sdp_e{email=Val} | Acc]);
parseLRDesc_1([#'PropertyParm'{name="p", value=[Val], extraInfo=EI}| PPs], 
	      Acc) ->
    parseLRDesc_1(PPs, [#sdp_p{phone_number=Val}  | Acc]);
parseLRDesc_1([P |PPs], Acc) ->
    parseLRDesc_1(PPs, Acc).

%%-----------------------------------------------------------------
%% Func: genLRDesc/2
%% Description:Generate the property groups of the LocalRemoteDescriptor
%%-----------------------------------------------------------------
genLRDesc([],_) ->
    [];
genLRDesc(asn1_NOVALUE, _) ->
    [];
genLRDesc([PropGrp |PropGrps], Coding) ->
    case genLRDesc_1(PropGrp, Coding, []) of
	no ->
	    genLRDesc(PropGrps, Coding);
	NewPropGrp ->
	    [NewPropGrp | genLRDesc(PropGrps, Coding)]
    end;
genLRDesc(Descr, Coding) 
  when record(Descr,'LocalRemoteDescriptor') ->
    DecodedPropGroups = genLRDesc(Descr#'LocalRemoteDescriptor'.propGrps, Coding),
    Descr#'LocalRemoteDescriptor'{propGrps = DecodedPropGroups}.


genLRDesc_1([], _, []) ->
    no;
genLRDesc_1([], _, Acc) ->
    lists:reverse(Acc);
genLRDesc_1([V| PPs], Coding, Acc) when record(V, sdp_v) ->
    PropP =  #'PropertyParm'{name="v", 
			     value=[
				    V#sdp_v.version
				   ]},
    genLRDesc_1(PPs, Coding, [PropP | Acc]);
genLRDesc_1([C| PPs], Coding, Acc) when record(C, sdp_c) ->
    PropP =  #'PropertyParm'{name="c", 
			     value=[
				    C#sdp_c.network_type ++ " " ++
				    C#sdp_c.address_type ++ " " ++
				    C#sdp_c.connection_addr
				   ]},
    genLRDesc_1(PPs, Coding, [PropP | Acc]);
genLRDesc_1([M| PPs], text, Acc) when record(M, sdp_m) ->
    NoOfPorts = case M#sdp_m.num_ports of
		    undefined ->
			"";
		    X ->
			"/" ++ integer_to_list(X) 
		end,
    PropP =  #'PropertyParm'{name="m", 
			     value=[
				    M#sdp_m.media ++ " " ++
				    integer_to_list(M#sdp_m.port) ++
				    NoOfPorts ++" " ++
				    M#sdp_m.transport ++ 
				    lists:foldl(fun(Element, String) -> 
							String ++ " " ++ Element 
						end, 
						[], 
						M#sdp_m.fmt_list)
				   ]},
    genLRDesc_1(PPs, text, [PropP | Acc]);
genLRDesc_1([M| PPs], binary, Acc) when record(M, sdp_m) ->
%    Media = case M#sdp_m.media of
%		"audio" -> 0;
%		"video" -> 1;
%		"data" -> 2;
%		Other ->
%		    exit({genLRDesc, "Media Type can't be binary encoded"})
%	    end,
%    Port
%    NoOfPorts = case M#sdp_m.num_ports of
%		    undefined ->
%			"";
%		    X ->
%			"/" ++ integer_to_list(X) 
%		end,
%    PropPs =  [#'PropertyParm'{name="Media", value=[Media]},
%     #'PropertyParm'{name="Port", value=[]},
%    #'PropertyParm'{name="PortType", value=[]},
%    #'PropertyParm'{name="NumChan", value=[]},
%    #'PropertyParm'{name="TransMode", value=[]},
%     #'PropertyParm'{name="SamplingRate", value=[]},
%     #'PropertyParm'{name="RTPpayload", value=[]},
%    #'PropertyParm'{name="IPv4", value=[]},
%     #'PropertyParm'{name="IPv6", value=[]},
%    #'PropertyParm'{name="UDP", value=[]},

%				    #sdp_m.name ++ " " ++
%				    integer_to_list(#sdp_m.port) ++
%				    NoOfPorts ++" " ++
%				    #sdp_m.address_type ++ " " ++
%				    #sdp_m.connection_addr
%				   ]}, 
    PropP=  #'PropertyParm'{},
    genLRDesc_1(PPs, binary, [PropP | Acc]);
genLRDesc_1([O| PPs], Coding, Acc) when record(O, sdp_o) ->
    PropP =  #'PropertyParm'{name="o", 
			     value=[
				    O#sdp_o.user_name ++ " " ++
				    O#sdp_o.session_id ++ " " ++
				    O#sdp_o.version ++ " " ++
				    O#sdp_o.network_type ++ " " ++
				    O#sdp_o.address_type ++ " " ++
				    O#sdp_o.address
				   ]},
    genLRDesc_1(PPs, Coding, [PropP | Acc]);
genLRDesc_1([A| PPs], Coding, Acc) when record(A, sdp_a_rtpmap) ->
    PropP =  #'PropertyParm'{name="a", 
			     value=["rtpmap:" ++
				    integer_to_list(A#sdp_a_rtpmap.payload_type) ++
				    " " ++
				    A#sdp_a_rtpmap.encoding_name ++ "/" ++
				    integer_to_list(A#sdp_a_rtpmap.clock_rate) ++
				    lists:foldl(fun(Element, String) -> 
							String ++ "/" ++ Element 
						end, 
						[], 
						A#sdp_a_rtpmap.encoding_parms)
				   ]},
    genLRDesc_1(PPs, Coding, [PropP | Acc]);
genLRDesc_1([A| PPs], Coding, Acc) when record(A, sdp_a_ptime) ->
    PropP =  #'PropertyParm'{name="a", 
			     value=["ptime:" ++
				    integer_to_list(A#sdp_a_ptime.packet_time)
				   ]},
    genLRDesc_1(PPs, Coding, [PropP | Acc]);
genLRDesc_1([A| PPs], Coding, Acc) when record(A, sdp_a) ->
    Value = case A#sdp_a.value of
		undefined ->
		    [];
		Val -> 
		    ":" ++ Val
	    end,
    PropP =  #'PropertyParm'{name="a", 
			     value=[
				    A#sdp_a.attribute ++ Value
				    ]},
    genLRDesc_1(PPs, Coding, [PropP | Acc]);

genLRDesc_1([B| PPs], Coding, Acc) when record(B, sdp_b) ->
    PropP =  #'PropertyParm'{name="b", 
			     value=[
				    B#sdp_b.modifier ++ ":" ++
				    integer_to_list(B#sdp_b.bandwidth)
				    ]},
    genLRDesc_1(PPs, Coding, [PropP | Acc]);
genLRDesc_1([T| PPs], Coding, Acc) when record(T, sdp_t) ->
    PropP =  #'PropertyParm'{name="t", 
			     value=[
				    integer_to_list(T#sdp_t.start) ++ 
				    " " ++
				    integer_to_list(T#sdp_t.stop)
				    ]},
    genLRDesc_1(PPs, Coding, [PropP | Acc]);
genLRDesc_1([R| PPs], Coding, Acc) when record(R, sdp_r) ->
    PropP =  #'PropertyParm'{name="r", 
			     value=[
				    R#sdp_r.repeat_interval ++ " " ++
				    R#sdp_r.active_duration ++ " " ++
				    lists:foldl(fun(Element, String) -> 
							String ++ " " ++ 
							    Element 
						end, 
						[], 
						R#sdp_r.list_of_offsets)
				   ]},
    genLRDesc_1(PPs, Coding, [PropP | Acc]);
genLRDesc_1([Z| PPs], Coding, Acc) when record(Z, sdp_z) ->
    PropP =  #'PropertyParm'{name="z", 
			     value=[
				    lists:foldl(fun(Element, String) -> 
							String ++ " " ++ 
							    Element 
						end, 
						[], 
						Z#sdp_z.list_of_adjustments)
				   ]},
    genLRDesc_1(PPs, Coding, [PropP | Acc]);
genLRDesc_1([K| PPs], Coding, Acc) when record(K, sdp_k) ->
    PropP =  #'PropertyParm'{name="k", 
			     value=[
				    K#sdp_k.method ++ ":" ++
				    K#sdp_k.encryption_key
				    ]},
    genLRDesc_1(PPs, Coding, [PropP | Acc]);

genLRDesc_1([S| PPs], Coding, Acc) when record(S, sdp_s) ->
    PropP =  #'PropertyParm'{name="s", 
			     value=[
				    S#sdp_s.name
				    ]},
    genLRDesc_1(PPs, Coding, [PropP | Acc]);
genLRDesc_1([I| PPs], Coding, Acc) when record(I, sdp_i) ->
    PropP =  #'PropertyParm'{name="i", 
			     value=[
				    I#sdp_i.session_descriptor
				    ]},
    genLRDesc_1(PPs, Coding, [PropP | Acc]);

genLRDesc_1([U| PPs], Coding, Acc) when record(U, sdp_u) ->
    PropP =  #'PropertyParm'{name="u", 
			     value=[
				    U#sdp_u.uri
				    ]},
    genLRDesc_1(PPs, Coding, [PropP | Acc]);

genLRDesc_1([E| PPs], Coding, Acc) when record(E, sdp_e) ->
    PropP =  #'PropertyParm'{name="e", 
			     value=[
				    E#sdp_e.email
				    ]},
    genLRDesc_1(PPs, Coding, [PropP | Acc]);

genLRDesc_1([P| PPs], Coding, Acc) when record(P, sdp_p) ->
    PropP =  #'PropertyParm'{name="p", 
			     value=[
				    P#sdp_p.phone_number
				    ]},
    genLRDesc_1(PPs, Coding, [PropP | Acc]);
genLRDesc_1([P |Ps], Coding, Acc) ->
    genLRDesc_1(Ps, Coding, Acc).
%%-----------------------------------------------------------------
%% Func: chooseSDPProps/2
%% Description: Check if a ephemeral termination is needed
%%-----------------------------------------------------------------
chooseSDPProps([], _) ->
    no;
chooseSDPProps(_, undefined) ->
    no;
chooseSDPProps(asn1_NOVALUE, Phonepid) ->
    V = #sdp_v{version = 0},
    C = #sdp_c{
	       network_type = "IN",
	       address_type = "IP4",
	       connection_addr = get_ip_from_pid_as_string(Phonepid) %??
	      },
    M = #sdp_m{  
	       media = "audio", 
	       port = 21000,
	       transport =  "RTP/AVP",
	       fmt_list = ["0"]
	      },
    [V,C,M];

%    #'LocalRemoteDescriptor'{propGrps = 
%			     [#'PropertyParm'{name="v", value=["0"]},
%			      #'PropertyParm'{name="c", value=[C]},
%			      #'PropertyParm'{name="c", value=[M]}]};
			       
chooseSDPProps([PropGrp |PropGrps], Phonepid) ->

    case chooseSDPProps_1(PropGrp, Phonepid, []) of
	no ->
	    chooseSDPProps(PropGrps, Phonepid);
	NewPropGrp ->
	    [NewPropGrp]
    end.

%chooseSDPProps(Desc, Phonepid) when record(Desc, 'LocalRemoteDescriptor') ->
%    #'LocalRemoteDescriptor'{propGrps = 
%			     chooseSDPProps(Desc#'LocalRemoteDescriptor'.propGrps, Phonepid)}.

chooseSDPProps_1([], _, []) ->
    no;
chooseSDPProps_1([], _, Acc) ->
    lists:reverse(Acc);
chooseSDPProps_1([#sdp_v{version=0} | PPs], Phonepid, Acc) ->
    chooseSDPProps_1(PPs, Phonepid, [#'sdp_v'{version=0} | Acc]);
chooseSDPProps_1([C | PPs],Phonepid, Acc) when record(C, sdp_c) ->
    case C of
	#sdp_c{network_type="IN", address_type = "IP4", connection_addr = [$$]} ->
	    chooseSDPProps_1(PPs, Phonepid, 
			     [C#sdp_c{connection_addr =
			       get_ip_from_pid_as_string(Phonepid)}
			      | Acc]);
	_ ->
	    no
    end;
chooseSDPProps_1([M | PPs], Phonepid, Acc) when record(M, sdp_m)->
    case M of
	#sdp_m{media = "audio", port = 0, num_ports = _,
	       transport = "RTP/AVP", fmt_list = _} ->
	    chooseSDPProps_1(PPs, Phonepid, [M#sdp_m{port = 21000} | Acc]);
	#sdp_m{media = "audio", port = _, num_ports = _,
	       transport = "RTP/AVP", fmt_list = _} ->
	    chooseSDPProps_1(PPs, Phonepid, [M | Acc]);
	_ ->
	    no
    end;
chooseSDPProps_1([A| PPs], Phonepid, Acc) when record(A, 'sdp_a')->
    chooseSDPProps_1(PPs, Phonepid, [A |Acc]);
chooseSDPProps_1([P |Ps], Phonepid, Acc) ->
    chooseSDPProps_1(Ps, Phonepid, Acc).

%%-----------------------------------------------------------------
%% Func: resetSDPPropsToChoose/1
%% Description: Change SDP props to unchoosen state 
%%-----------------------------------------------------------------

resetSDPPropsToChoose([]) ->
    [];
resetSDPPropsToChoose([PropGrp |PropGrps]) ->
    [resetSDPPropsToChoose_1(PropGrp, []) |
     resetSDPPropsToChoose(PropGrps)].

resetSDPPropsToChoose_1([], Acc) ->
    lists:reverse(Acc);
resetSDPPropsToChoose_1([#sdp_v{version=0} | PPs], Acc)  ->
    resetSDPPropsToChoose_1(PPs, [#sdp_v{version=0} | Acc]);
resetSDPPropsToChoose_1([C| PPs], Acc) when record(C, sdp_c) ->
    case C of
	#sdp_c{network_type="IN", address_type = "IP4", connection_addr = _} ->
	    resetSDPPropsToChoose_1(PPs, 
				    [C#sdp_c{connection_addr = [$$]} 
				     | Acc]);
	_ ->
	    no
    end;
resetSDPPropsToChoose_1([M| PPs], Acc) when record(M, sdp_m) ->
    case M of
	#sdp_m{media = "audio", port = _, num_ports = _, 
	       transport = "RTP/AVP", fmt_list = _}	->
	    resetSDPPropsToChoose_1(PPs, [M#sdp_m{port = 0} | Acc]);
	_ ->
	    no
    end;
resetSDPPropsToChoose_1([A | PPs], Acc) when record(A, sdp_a) ->
    resetSDPPropsToChoose_1(PPs, [A |Acc]);
resetSDPPropsToChoose_1([P |Ps], Acc) ->
    resetSDPPropsToChoose_1(Ps, [P |Acc]).


%%-----------------------------------------------------------------
%% Func: connection_addr/1
%% Description: Get IP address from SDP descriptor
%%-----------------------------------------------------------------

connection_addr([PropGrp]) ->
    connection_addr_1(PropGrp);
connection_addr(X) ->
    exit("No or more than one property group in "
	 "choosen LocalRemoteDescriptor").


connection_addr_1([C| PPs]) when record(C, sdp_c)  ->
	C#sdp_c.connection_addr;
connection_addr_1([P |Ps]) ->
    connection_addr_1(Ps).

%%-----------------------------------------------------------------
%% Func: port/1
%% Description: Get RTP port from SDP descriptor
%%-----------------------------------------------------------------

port([PropGrp]) ->
    port_1(PropGrp);
port(X) ->
    exit("No or more than one property group in "
	 "choosen LocalRemoteDescriptor").

port_1([M| PPs])  when record(M, sdp_m) ->
	M#sdp_m.port;
port_1([P |Ps]) ->
    port_1(Ps).

%%-----------------------------------------------------------------
%% Func: payload_type/1
%% Description: Change SDP props to ...
%%-----------------------------------------------------------------

payload_type([PropGrp]) ->
    payload_type_1(PropGrp);
payload_type(X) ->
    exit("No or more than one property group in "
	 "choosen LocalRemoteDescriptor").

payload_type_1([M | PPs]) ->
    M#sdp_m.fmt_list;
payload_type_1([P |Ps]) ->
    payload_type_1(Ps).

%%======================================================================
%% Internal functions
%%======================================================================

%% These two functions are copied from the "utils.erl" module in the 
%% example directory. This _is_ a temporary solution since this module
%% should be rewritten anyway...
%%
get_ip_from_pid_as_string(Pid) ->
    [_, Host] = string:tokens(atom_to_list(node(Pid)), "@"),
    {ok, {A1, A2, A3, A4}} = inet:getaddr(Host, inet),
    ip_address_as_string({A1, A2, A3, A4}).

ip_address_as_string({A1, A2, A3, A4}) ->
    integer_to_list(A1) ++ "." ++ 
        integer_to_list(A2) ++ "." ++ 
        integer_to_list(A3) ++ "." ++ 
        integer_to_list(A4).
